#include  "built-in-types.h"
#include  "global.h"
#include  "constants.h"
#include  <sstream>
#include  <cctype>
#include  "basic_io.h"
#include  "prerootdata.h"
#include  "dynkin.h"
#include  "arithmetic.h"
#include  "lattice.h"
#include  "matrix.h"
#include  "matreduc.h"
#include  "tags.h"
#include  "tori.h"
#include  "weyl.h"
#include  "dynkin.h"
#include  "poset.h"
#include  "prettyprint.h"
#include  "ioutils.h"
#include  <iomanip> 
#include  "evaluator.h" 
#include  "realredgp_io.h"
#include  "blocks.h"
#include  "block_io.h"
#include  "kgb.h"
#include  "kgb_io.h"
#include  "kl.h"
#include  "klsupport.h"
#include  "kl_io.h"
#include  "wgraph.h"
#include  "wgraph_io.h"

namespace atlas { namespace interpreter {
namespace { 
            inline std::istream& skip_punctuation(std::istream &is)
            { char c; do is>>c; while (is and std::ispunct(c));
                return is.unget();
            }
            
            void Lie_type_wrapper(expression_base::level l)
               throw(std::bad_alloc,std::runtime_error)
            { std::istringstream is(get<string_value>()->val);
              own_Lie_type result = std::make_shared<Lie_type_value>();
              size_t total_rank=0; char c;
              while (skip_punctuation(is)>>c) // i.e., until |not is.good()|
              { size_t rank;
                if (is.peek()=='-' or not (is>>rank)) // explicitly forbid minus sign
                  throw std::runtime_error
                   ("Error in type string '"+is.str()+"' for Lie type");
            
                result->add_simple_factor(c,rank);
                  // this may |throw| a |runtime_error| as well
                if ((total_rank+=rank)>constants::RANK_MAX)
                  throw std::runtime_error
                  ("Total rank exceeds implementation limit "+str(constants::RANK_MAX));
            
              }
              if (l!=expression_base::no_value)
                push_value(std::move(result));
            }
            
            void Lie_type_coercion()
            { Lie_type_wrapper(expression_base::single_value); }
            
            void Cartan_matrix_wrapper(expression_base::level l)
            { shared_Lie_type t=get<Lie_type_value>();
              if (l==expression_base::no_value)
                return;
              push_value(std::make_shared<matrix_value>(t->val.Cartan_matrix()));
            }
            
            void type_of_Cartan_matrix_wrapper (expression_base::level l)
            { shared_matrix m=get<matrix_value>();
              Permutation pi;
              LieType lt=dynkin::Lie_type(m->val,true,true,pi);
              if (l==expression_base::no_value)
                return;
              push_value(std::make_shared<Lie_type_value>(lt));
              push_value(std::make_shared<vector_value>(CoeffList(pi.begin(),pi.end())));
              if (l==expression_base::single_value)
                wrap_tuple<2>();
            }
            
            void Lie_rank_wrapper(expression_base::level l)
            { shared_Lie_type t=get<Lie_type_value>();
              if (l!=expression_base::no_value)
                push_value(std::make_shared<int_value>(t->val.rank()));
            }
            void semisimple_rank_wrapper(expression_base::level l)
            { shared_Lie_type t=get<Lie_type_value>();
              if (l!=expression_base::no_value)
                push_value(std::make_shared<int_value>(t->val.semisimple_rank()));
            }
            void Lie_type_string_wrapper(expression_base::level l)
            { std::ostringstream s; s << get<Lie_type_value>()->val;
              if (l!=expression_base::no_value)
                push_value(std::make_shared<string_value>(s.str()));
            }
            
            void nr_factors_wrapper(expression_base::level l)
            { shared_Lie_type t=get<Lie_type_value>();
              if (l!=expression_base::no_value)
                push_value(std::make_shared<int_value>(t->val.size()));
            }
            
            void Lie_factors_wrapper(expression_base::level l)
            { shared_Lie_type t=get<Lie_type_value>();
              if (l==expression_base::no_value)
                return;
              own_row result = std::make_shared<row_value>(t->val.size());
              for (unsigned i=0; i<t->val.size(); ++i)
              { std::vector<SimpleLieType> factor(1,t->val[i]);
                result->val[i]=std::make_shared<Lie_type_value>(LieType(factor));
              }
              push_value(std::move(result));
            }
            
            void Smith_Cartan_wrapper(expression_base::level l)
            { shared_Lie_type t=get<Lie_type_value>();
              if (l==expression_base::no_value)
                return;
              own_vector inv_factors = std::make_shared<vector_value>(CoeffList());
              WeightList b = t->val.Smith_basis(inv_factors->val);
              push_value(std::make_shared<matrix_value>(LatticeMatrix(b,b.size())));
              push_value(std::move(inv_factors));
              if (l==expression_base::single_value)
                 wrap_tuple<2>();
            }
            
            void filter_units_wrapper (expression_base::level l)
            { own_vector inv_f=get_own<vector_value>();
              own_matrix basis=get_own<matrix_value>();
              unsigned int n=basis->val.numColumns();
              if (inv_f->val.size()>n)
                throw std::runtime_error ("Too many factors: "+
            
                  str(inv_f->val.size()) + " for " +str(n)+ " columns");
              if (l==expression_base::no_value)
                return;
            
              size_t i=0;
              while (i<n)
                if (i>=inv_f->val.size() or inv_f->val[i]!=1)
                  ++i; // keep invariant factor and column
                else
                { inv_f->val.erase(inv_f->val.begin()+i);
                    basis->val.eraseColumn(i);
                }
              push_value(basis); push_value(inv_f);
              if (l==expression_base::single_value)
                  wrap_tuple<2>();
            }
            
            LatticeMatrix 
            annihilator_modulo(const LatticeMatrix& M, arithmetic::Denom_t denominator)
            
            { int_Matrix row,col;
              CoeffList lambda = matreduc::diagonalise(M,row,col);
            
              for (size_t i=0; i<lambda.size(); ++i)
                row.rowMultiply(i,arithmetic::div_gcd(denominator,lambda[i]));
            
              return row.transposed();
            }
            
            void ann_mod_wrapper(expression_base::level l)
            { shared_int d=get<int_value>();
              shared_matrix m=get<matrix_value>();
            
              if (l!=expression_base::no_value)
                push_value(std::make_shared<matrix_value>(annihilator_modulo(m->val,d->val)));
            }
            
            void replace_gen_wrapper (expression_base::level l)
            { shared_matrix new_generators=get<matrix_value>();
              push_tuple_components(); // a pair as returned by \.{Smith\_Cartan}
              shared_vector inv_f=get<vector_value>();
              own_matrix generators=get_own<matrix_value>();
              unsigned int n=generators->val.numColumns();
              if (inv_f->val.size()>n)
                throw std::runtime_error ("Too many factors: "+
            
                  str(inv_f->val.size()) + " for " +str(n)+ " columns");
            
              if (new_generators->val.numRows()!=generators->val.numRows())
                throw std::runtime_error("Column lengths do not match");
            
            
              size_t k=0; // index to replacement generators
              for (size_t j=0; j<n; ++j)
                if (j>=inv_f->val.size() or inv_f->val[j]!=1)
                   // replace column |j| by |k| from |new_generators|
                { if (k>=new_generators->val.numColumns())
                    throw std::runtime_error ("Not enough replacement columns");
            
                  generators->val.set_column(j,new_generators->val.column(k));
                  ++k;
                }
              if (k<new_generators->val.numColumns())
                    throw std::runtime_error ("Too many replacement columns");
            
              if (l!=expression_base::no_value)
                push_value(generators);
            }
            
            void quotient_basis_wrapper(expression_base::level l)
            { shared_row L=get<row_value>();
              // and leave Lie type on stack for $Smith\_Cartan$
              Smith_Cartan_wrapper(expression_base::multi_value);
              shared_value SC_basis = *(execution_stack.end()-2);
              shared_value invf = *(execution_stack.end()-1);
              wrap_tuple<2>(); // |replace_gen| wants as first argument a tuple
            push_value(SC_basis);
              push_value(invf);
              filter_units_wrapper(expression_base::multi_value);
              shared_vector v=get<vector_value>();
              // and leave $C$ for call to $mm\_prod$
            
              LatticeMatrix M(v->val.size(),L->length());
              arithmetic::Denom_t d=1;
              
              { std::vector<arithmetic::Denom_t> denom(L->length());
                for (size_t j=0; j<L->length(); ++j)
                { const RatWeight& gen =
                    force<rational_vector_value>(&*L->val[j])->val;
                  denom[j] = gen.denominator();
                  d=arithmetic::lcm(d,denom[j]);
              
                  if (gen.size()!=v->val.size())
                    throw std::runtime_error
                      ("Length mismatch for generator "+str(j) +": "
              
                      +str(gen.size()) + ':' + str(v->val.size()));
              
                  Weight col(gen.numerator().begin(),gen.numerator().end()); // convert
                  for (size_t i=0; i<v->val.size(); ++i)
                  { if (v->val[i]*col[i]%gen.denominator()!=0) // must use signed arithmetic!!
              	throw std::runtime_error("Improper generator entry: "
              
                       +str(col[i])+'/'+str(denom[j])+" not a multiple of 1/"
                       +str(v->val[i]));
                    M(i,j) = arithmetic::remainder(col[i],denom[j]);
                    // ``mod $\Z$''; makes |M(i,j)| non-negative
                  }
                }
              // loop must end and restart here so that computation of |d| will be complete
              
                for (size_t j=0; j<L->length(); ++j) // convert to common denominator |d|
                { arithmetic::Denom_t f=d/denom[j];
                  for (size_t i=0; i<v->val.size(); ++i)
                    M(i,j) *= f;
                }
              }
              push_value(std::make_shared<matrix_value>(annihilator_modulo(M,d)));
            mm_prod_wrapper(expression_base::single_value);
            replace_gen_wrapper(l); // pass level parameter to final call
            }
            
            InnerClassType transform_inner_class_type
              (const char* s, const LieType& lt)
            throw (std::bad_alloc, std::runtime_error)
            { static const std::string types(lietype::innerClassLetters);
                // |"Ccesu"|
              InnerClassType result; // initially empty
              std::istringstream is(s);
              char c;
              size_t i=0; // position in simple factors of Lie type |lt|
              while (skip_punctuation(is)>>c)
                
                { if (types.find(c) == std::string::npos) throw std::runtime_error
                    (std::string("Unknown inner class symbol `")+c+"'");
                  if (i>= lt.size()) throw std::runtime_error("Too many inner class symbols");
                
                  lietype::TypeLetter t = lt[i].type(); size_t r=lt[i].rank();
                  if (c=='C') // complex inner class, which requires two equal simple factors
                  { if (i+1>=lt.size() or lt[i+1]!=lt[i]) throw std::runtime_error 
                      ("Complex inner class needs two identical consecutive types");
                
                    result.push_back('C');
                    i+=2; // advance past both simple factors
                  }
                  else if (c=='c' or c=='e')
                  { result.push_back('c'); ++i; } // synonyms
                  else if (c=='s')
                  { 
                    if (t=='A' and r>=2 or
                        t=='D' and r%2!=0 or
                        t=='E' and r==6 or
                        t=='T')
                      result.push_back('s');
                    else result.push_back('c');
                    ++i;
                  }
                  else if (c=='u') // unequal rank, the only remaining possibility
                  { 
                    { if (t=='D') result.push_back(r%2==0 ? 'u' : 's');
                      else if (t=='A' and r>=2 or t=='E' and r==6 or t=='T') result.push_back('s');
                      else throw std::runtime_error 
                        (std::string("Unequal rank class is meaningless for type ")+t+str(r));
                    
                    }
                    ++i;
                  }
                }
              if (i< lt.size()) throw std::runtime_error("Too few inner class symbols");
            
              return result;
            }
            
            void basic_involution_wrapper(expression_base::level l)
            { shared_string str=get<string_value>();
              shared_Lie_type t=get<Lie_type_value>();
            push_value(std::make_shared<matrix_value>  (lietype::involution
                       (t->val,transform_inner_class_type(str->val.c_str(),t->val))));
            }
            
            void based_involution_wrapper(expression_base::level l)
            { shared_string s = get<string_value>();
            shared_matrix basis = get<matrix_value>();
            shared_Lie_type type = get<Lie_type_value>();
            
              size_t r=type->val.rank();
              if (basis->val.numRows()!=r or basis->val.numRows()!=r)
                throw std::runtime_error 
                ("Basis should be given by "+str(r)+'x'+str(r)+" matrix");
            
            
              WeightInvolution inv=lietype::involution
                    (type->val,transform_inner_class_type(s->val.c_str(),type->val));
              try
              { push_value(std::make_shared<matrix_value>(inv.on_basis(basis->val.columns()))); }
              catch (std::runtime_error&) // relabel |"Inexact integer division"|
              { throw std::runtime_error
                  ("Inner class is not compatible with given lattice");
            
              }
            }
            
            LieType type_of_datum(const RootDatum& rd)
            { int_Matrix Cartan = rd.cartanMatrix();
            LieType t = dynkin::Lie_type(Cartan);
              if (!rd.isSemisimple())
                for (size_t i=rd.semisimpleRank(); i<rd.rank(); ++i)
                  t.push_back(SimpleLieType('T',1));
              return t;
            }
            
            void type_of_root_datum_wrapper(expression_base::level l)
            { shared_root_datum rd(get<root_datum_value>());
              push_value(std::make_shared<Lie_type_value>(type_of_datum(rd->val)));
            }
            
            void raw_root_datum_wrapper(expression_base::level l)
            { shared_matrix simple_coroots=get<matrix_value>();
              shared_matrix simple_roots=get<matrix_value>();
            
              size_t nr = simple_roots->val.numRows(),
                     nc = simple_roots->val.numColumns();
            
              if (simple_coroots->val.numRows()!=nr  or
                  simple_coroots->val.numColumns()!=nc)
                throw std::runtime_error
                ("Sizes (" +str(nr)+ "," +str(nc) +"),("+
                  str(simple_coroots->val.numRows()) +","+
                  str(simple_coroots->val.numColumns()) +
                  ") of simple (co)root systems differ");
            
            
              WeightList s; CoweightList c;
              s.reserve(nc);
              c.reserve(nc);
            
              for (size_t j=0; j<nc; ++j)
            { s.push_back(simple_roots->val.column(j));
                  c.push_back(simple_coroots->val.column(j));
              }
            
              PreRootDatum prd(s,c,nr);
              try { Permutation dummy;
                dynkin::Lie_type(prd.Cartan_matrix(),true,true,dummy);
              }
              catch (error::CartanError)
            {
                throw std::runtime_error("System of (co)roots has invalid Cartan matrix");
            }
            
              if (l!=expression_base::no_value)
                push_value(std::make_shared<root_datum_value>  (RootDatum(prd)));
            }
            
            void root_datum_wrapper(expression_base::level l)
            { shared_matrix lattice=get<matrix_value>();
              shared_Lie_type type=get<Lie_type_value>();
              if (lattice->val.numRows()!=lattice->val.numColumns()  or
                  lattice->val.numRows()!=type->val.rank())
                throw std::runtime_error
                ("Sub-lattice matrix should have size " 
            
                  +str(type->val.rank())+'x'+str(type->val.rank()));
              PreRootDatum prd(type->val,lattice->val.columns());
            
            
            
              if (l!=expression_base::no_value)
                push_value(std::make_shared<root_datum_value>  (RootDatum(prd)));
            }
            
            void sublattice_root_datum_wrapper(expression_base::level l)
            { shared_matrix lattice=get<matrix_value>();
              shared_root_datum rd=get<root_datum_value>();
              if (lattice->val.numRows()!=lattice->val.numColumns()  or
                  lattice->val.numRows()!=rd->val.rank())
                throw std::runtime_error
                ("Sub-lattice matrix should have size " 
            
                  +str(rd->val.rank())+'x'+str(rd->val.rank()));
            
              WeightList simple_roots(rd->val.beginSimpleRoot(),rd->val.endSimpleRoot());
              CoweightList simple_coroots
                    (rd->val.beginSimpleCoroot(),rd->val.endSimpleCoroot());
              PreRootDatum prd(simple_roots,simple_coroots,rd->val.rank());
              prd.quotient(lattice->val);
            
            
            
              if (l!=expression_base::no_value)
                push_value(std::make_shared<root_datum_value>  (RootDatum(prd)));
            }
            
            void quotient_datum_wrapper(expression_base::level l)
            { shared_value L = pop_value();
              shared_value lt= pop_value();
              push_value(lt); // the Lie type, for call of $root\_datum$
            push_value(lt);
              push_value(L); quotient_basis_wrapper(expression_base::single_value);
            root_datum_wrapper(l); // pass level parameter to final call
            }
            
            void simply_connected_datum_wrapper(expression_base::level l)
            { if (l==expression_base::no_value)
              { execution_stack.pop_back();
                return; // no possibilities of errors, so avoid useless work
              }
              size_t rank =
                force<Lie_type_value>(execution_stack.back().get())->val.rank();
              push_value(std::make_shared<int_value>(rank));
              id_mat_wrapper(expression_base::single_value);
            root_datum_wrapper(expression_base::single_value);
            }
            
            void adjoint_datum_wrapper(expression_base::level l)
            { if (l==expression_base::no_value)
              { execution_stack.pop_back();
                return; // no possibilities of errors, so avoid useless work
              }
              push_value(execution_stack.back()); // duplicate Lie type argument
              Cartan_matrix_wrapper(expression_base::single_value);
              transpose_mat_wrapper(expression_base::single_value);
              own_matrix M=get_own<matrix_value>();
              for (size_t i=0; i<M->val.numRows(); ++i)
                if (M->val(i,i)==0) M->val(i,i)=1;
              push_value(M);
            root_datum_wrapper(expression_base::single_value);
            }
            
            void simple_roots_wrapper(expression_base::level l)
            { shared_root_datum rd(get<root_datum_value>());
              if (l==expression_base::no_value)
                return;
              WeightList srl
                (rd->val.beginSimpleRoot(),rd->val.endSimpleRoot());
              push_value(std::make_shared<matrix_value>(int_Matrix(srl,rd->val.rank())));
            }
            
            void simple_coroots_wrapper(expression_base::level l)
            { shared_root_datum rd(get<root_datum_value>());
              if (l==expression_base::no_value)
                return;
              WeightList scl
                (rd->val.beginSimpleCoroot(),rd->val.endSimpleCoroot());
              push_value(std::make_shared<matrix_value>(int_Matrix(scl,rd->val.rank())));
            }
            
            void datum_Cartan_wrapper(expression_base::level l)
            { shared_root_datum rd(get<root_datum_value>());
              if (l==expression_base::no_value)
                return;
              int_Matrix M = rd->val.cartanMatrix();
              push_value(std::make_shared<matrix_value>(M));
            }
            
            void positive_roots_wrapper(expression_base::level l)
            { shared_root_datum rd(get<root_datum_value>());
              if (l==expression_base::no_value)
                return;
              WeightList rl
                (rd->val.beginPosRoot(),rd->val.endPosRoot());
              push_value(std::make_shared<matrix_value>
                (int_Matrix(rl,rd->val.rank())));
            }
            
            void positive_coroots_wrapper(expression_base::level l)
            { shared_root_datum rd(get<root_datum_value>());
              if (l==expression_base::no_value)
                return;
              WeightList crl
                (rd->val.beginPosCoroot(),rd->val.endPosCoroot());
              push_value(std::make_shared<matrix_value>
                (int_Matrix(crl,rd->val.rank())));
            }
            void roots_wrapper(expression_base::level l)
            { shared_root_datum rd(get<root_datum_value>());
              if (l==expression_base::no_value)
                return;
              WeightList rl
                (rd->val.beginRoot(),rd->val.endRoot());
              push_value(std::make_shared<matrix_value>(int_Matrix(rl,rd->val.rank())));
            }
            
            void coroots_wrapper(expression_base::level l)
            { shared_root_datum rd(get<root_datum_value>());
              if (l==expression_base::no_value)
                return;
              WeightList crl
                (rd->val.beginCoroot(),rd->val.endCoroot());
              push_value(std::make_shared<matrix_value>(int_Matrix(crl,rd->val.rank())));
            }
            
            void rd_rank_wrapper(expression_base::level l)
            { shared_root_datum rd(get<root_datum_value>());
              if (l!=expression_base::no_value)
                push_value(std::make_shared<int_value>(rd->val.rank()));
            }
            
            void rd_semisimple_rank_wrapper(expression_base::level l)
            { shared_root_datum rd(get<root_datum_value>());
              if (l!=expression_base::no_value)
                push_value(std::make_shared<int_value>(rd->val.semisimpleRank()));
            }
            
            void rd_nposroots_wrapper(expression_base::level l)
            { shared_root_datum rd(get<root_datum_value>());
              if (l!=expression_base::no_value)
                push_value(std::make_shared<int_value>(rd->val.numPosRoots()));
            }
            
            void root_coradical_wrapper(expression_base::level l)
            { shared_root_datum rd(get<root_datum_value>());
              if (l==expression_base::no_value)
                return;
              WeightList srl
                (rd->val.beginSimpleRoot(),rd->val.endSimpleRoot());
              srl.insert(srl.end(),rd->val.beginCoradical(),rd->val.endCoradical());
              push_value(std::make_shared<matrix_value>(int_Matrix(srl,rd->val.rank())));
            }
            
            void coroot_radical_wrapper(expression_base::level l)
            { shared_root_datum rd(get<root_datum_value>());
              if (l==expression_base::no_value)
                return;
              WeightList scl
                (rd->val.beginSimpleCoroot(),rd->val.endSimpleCoroot());
              scl.insert(scl.end(),rd->val.beginRadical(),rd->val.endRadical());
              push_value(std::make_shared<matrix_value>(int_Matrix(scl,rd->val.rank())));
            }
            
            void fundamental_weight_wrapper(expression_base::level l)
            { int i= get<int_value>()->val;
              shared_root_datum rd(get<root_datum_value>());
              if (unsigned(i)>=rd->val.semisimpleRank())
                throw std::runtime_error("Invalid index "+str(i));
              if (l!=expression_base::no_value)
                push_value(std::make_shared<rational_vector_value>
                  (rd->val.fundamental_weight(i).normalize()));
            }
            
            void fundamental_coweight_wrapper(expression_base::level l)
            { int i= get<int_value>()->val;
              shared_root_datum rd(get<root_datum_value>());
              if (unsigned(i)>=rd->val.semisimpleRank())
                throw std::runtime_error("Invalid index "+str(i));
              if (l!=expression_base::no_value)
                push_value(std::make_shared<rational_vector_value>
                  (rd->val.fundamental_coweight(i).normalize()));
            }
            
            void dual_datum_wrapper(expression_base::level l)
            { shared_root_datum rd(get<root_datum_value>());
              if (l!=expression_base::no_value)
                push_value(std::make_shared<root_datum_value>
                  (RootDatum(rd->val,tags::DualTag())));
            }
            
            void derived_info_wrapper(expression_base::level l)
            { shared_root_datum rd(get<root_datum_value>());
              if (l!=expression_base::no_value)
              { int_Matrix projector;
                push_value(std::make_shared<root_datum_value>
                  (RootDatum(projector,rd->val,tags::DerivedTag())));
                push_value(std::make_shared<matrix_value>(projector));
                if (l==expression_base::single_value)
                  wrap_tuple<2>();
              }
            }
            
            void mod_central_torus_info_wrapper(expression_base::level l)
            { shared_root_datum rd(get<root_datum_value>());
              if (l!=expression_base::no_value)
              { int_Matrix injector;
                push_value(std::make_shared<root_datum_value>
                  (RootDatum(injector,rd->val,tags::AdjointTag())));
                push_value(std::make_shared<matrix_value>(injector));
                if (l==expression_base::single_value)
                  wrap_tuple<2>();
              }
            }
            
            void integrality_datum_wrapper(expression_base::level l)
            { shared_rational_vector lambda = get<rational_vector_value>();
              shared_root_datum rd(get<root_datum_value>());
              if (lambda->val.size()!=rd->val.rank())
              { std::ostringstream o;
                o << "Length " << lambda->val.size()
                  << " of rational vector differs from rank " << rd->val.rank();
                throw std::runtime_error(o.str());
            
              }
              if (l!=expression_base::no_value)
                push_value(std::make_shared<root_datum_value> 
                  (rootdata::integrality_datum(rd->val,lambda->val)));
            }
            
            void integrality_points_wrapper(expression_base::level l)
            { shared_rational_vector lambda = get<rational_vector_value>();
              shared_root_datum rd(get<root_datum_value>());
              if (lambda->val.size()!=rd->val.rank())
              { std::ostringstream o;
                o << "Length " << lambda->val.size()
                  << " of rational vector differs from rank " << rd->val.rank();
                throw std::runtime_error(o.str());
            
              }
              if (l==expression_base::no_value)
                return;
            
              RationalList ipl = rootdata::integrality_points(rd->val,lambda->val);
                // method normalises rationals
              own_row result = std::make_shared<row_value>(ipl.size());
              for (size_t i=0; i<ipl.size(); ++i)
                result->val[i]=std::make_shared<rat_value>(ipl[i]);
              push_value(std::move(result));
            }
            
            std::pair<size_t,size_t> classify_involution
              (const WeightInvolution& M)
            throw (std::bad_alloc, std::runtime_error)
            { size_t r=M.numRows();
              
              { if (M.numRows()!=r or M.numColumns()!=r) throw std::runtime_error
                  ("Involution should be a "+str(r)+"x"+str(r)+" matrix, got a "
              
                   +str(M.numRows())+"x"+str(M.numColumns())+" matrix");
                WeightInvolution Id(r),Q(M*M);
                if (Q!=Id) throw std::runtime_error
                    ("Given transformation is not an involution");
              
              }
              tori::RealTorus Tor(M);
              return std::make_pair(Tor.compactRank(),Tor.splitRank());
            }
            
            void classify_wrapper(expression_base::level l)
            { shared_matrix M(get<matrix_value>());
              std::pair<size_t,size_t> p=classify_involution(M->val);
              if (l==expression_base::no_value)
                return;
              push_value(std::make_shared<int_value>(p.first)); // compact rank
              push_value(std::make_shared<int_value>((M->val.numRows()-p.first-p.second)/2));
                // Complex rank
              push_value(std::make_shared<int_value>(p.second)); // split rank
              if (l==expression_base::single_value)
                wrap_tuple<3>();
            }
            
            lietype::Layout check_involution
             (const WeightInvolution& M, const RootDatum& rd,
              WeylWord& ww)
             throw (std::bad_alloc, std::runtime_error)
            { size_t r=rd.rank(),s=rd.semisimpleRank();
              
              { if (M.numRows()!=r or M.numColumns()!=r) throw std::runtime_error
                  ("Involution should be a "+str(r)+"x"+str(r)+" matrix, got a "
              
                   +str(M.numRows())+"x"+str(M.numColumns())+" matrix");
                WeightInvolution Id(r),Q(M*M);
                if (Q!=Id) throw std::runtime_error
                    ("Given transformation is not an involution");
              
              }
            Permutation p(s);
              
              { RootNbrList Delta(s);
                for (weyl::Generator i=0; i<s; ++i)
                { Delta[i]=rd.rootNbr(M*rd.simpleRoot(i));
                  if (Delta[i]==rd.numRoots()) // then image not found
                    throw std::runtime_error
                      ("Matrix maps simple root "+str(i)+" to non-root");
                }
                ww = wrt_distinguished(rd,Delta);
                for (weyl::Generator i=0; i<s; ++i)
                  if (rd.isSimpleRoot(Delta[i])) // should have made every root simple
                    p[i]=rd.simpleRootIndex(Delta[i]);
                  else throw std::runtime_error
                    ("Matrix does not define a root datum automorphism");
              
                for (weyl::Generator i=0; i<s; ++i)
                  for (weyl::Generator j=0; j<s; ++j)
                    if (rd.cartan(p[i],p[j])!=rd.cartan(i,j)) throw std::runtime_error
                    ("Matrix does not define a root datum automorphism");
              }
            lietype::Layout result;
            LieType& type=result.d_type;
              InnerClassType& inner_class=result.d_inner;
              Permutation& pi=result.d_perm;
              
              { int_Matrix Cartan = rd.cartanMatrix();
                RankFlagsList comp =
                  dynkin::components(dynkin::DynkinDiagram(Cartan));
              
                type.reserve(comp.size()+r-s);
                inner_class.reserve(comp.size()+r-s); // certainly enough
                type = dynkin::Lie_type(Cartan,true,false,pi);
                  // no need to check validity of |Cartan|
                assert(type.size()==comp.size());
                size_t offset=0; // accumulated rank of simple factors seen
              
                for (size_t i=0; i<type.size(); ++i)
                { bool equal_rank=true;
                  size_t comp_rank = type[i].rank();
                  assert (comp_rank==comp[i].count());
                     // and |pi[j]| runs through |comp[i]| in following loop
                  for (size_t j=offset; j<offset+comp_rank; ++j) // traverse component
                    if (p[pi[j]]!=pi[j]) { equal_rank=false; break; }
                if (equal_rank) inner_class.push_back('c');
                    // identity on this component: compact component
                  else if(comp[i].test(p[pi[offset]]))
                     // (any) one root stays in |comp[i]|; unequal rank
                    inner_class.push_back(type[i].first=='D' and comp_rank%2==0 ? 'u' : 's');
                  else
                  { inner_class.push_back('C'); // record Complex component
                    
                    { size_t beta = p[pi[offset]];
                      
                      { size_t j, k;
                        for (j=offset+comp_rank,k=i+1; k<type.size(); ++k)
                          if (comp[k].test(beta))
                            break;
                          else
                            j+=type[k].second;
                        if (k==type.size())
                          throw std::logic_error("Non matching Complex factor");
                      
                      
                      #ifndef NDEBUG
                        assert(type[k]==type[i]); // paired simple types for complex factor
                        for (size_t l=1; l<comp_rank; ++l)
                          assert(comp[k].test(p[pi[offset+l]]));
                              // image by |p| of remainder of |comp[i]| matches |comp[k]|
                      #endif
                      
                        if (k>i+1) // then we need to move component |k| down to |i+1|
                        {
                          while (j-->offset+comp_rank) // shift up intermediate components in |pi|
                            pi[j+comp_rank]=pi[j];
                      
                          RankFlags match_comp=comp[k];
                            // save component matching |comp[i]| under |p|
                      
                          while(k-->i+1) // shift intermediate simple factors
                          { type[k+1]=type[k]; comp[k+1]=comp[k]; }
                      
                          comp[i+1]=match_comp; // reinsert mathcing component after |comp[i]|
                        }
                      
                      }
                    
                      type[i+1]=type[i]; // duplicate factor |i|
                      for (size_t j=offset; j<offset+comp_rank; ++j)
                        pi[j+comp_rank]=p[pi[j]];
                         // reconstruct matching component in matching order
                    }
                    offset += comp_rank;
                    ++i; // skip over component |i|, loop will skip component |i+1|
                  }
              
                  offset += comp_rank;
                } // |for (i)|
              }
              if (r>s)
                
                { for (size_t k=0; k<r-s; ++k)
                    type.push_back(SimpleLieType('T',1));
                  int_Matrix root_lattice
                    (rd.beginSimpleRoot(),rd.endSimpleRoot(),r,tags::IteratorTag());
                CoeffList factor; // values will be unused
                  int_Matrix basis =
                     matreduc::adapted_basis(root_lattice,factor);
                WeightInvolution inv =
                     basis.inverse().block(s,0,r,r)*M*basis.block(0,s,r,r);
                    // involution on quotient by root lattice
                  std::pair<size_t,size_t> cl=classify_involution(inv);
                size_t& compact_rank=cl.first;
                  size_t& split_rank=cl.second;
                  size_t Complex_rank=(r-s-compact_rank-split_rank)/2;
                while (compact_rank-->0) inner_class.push_back('c');
                  while (Complex_rank-->0) inner_class.push_back('C');
                  while (split_rank-->0) inner_class.push_back('s');
                }
              return result;
            }
            
            void fix_involution_wrapper(expression_base::level l)
            { LatticeMatrix M(get<matrix_value>()->val); // safe use of temporary
              shared_root_datum rd(get<root_datum_value>());
              WeylWord ww;
              lietype::Layout lo = check_involution(M,rd->val,ww);
              if (l==expression_base::no_value)
                return;
            
              for (unsigned int i=0; i<ww.size(); ++i) // apply elements in generation order
                rd->val.simple_reflect(ww[i],M);
              std::unique_ptr<ComplexReductiveGroup>
                G(new ComplexReductiveGroup(rd->val,M));
              push_value(std::make_shared<inner_class_value>(std::move(G),lo));
            }
            
            void twisted_involution_wrapper(expression_base::level l)
            { LatticeMatrix M(get<matrix_value>()->val);
              shared_root_datum rd(get<root_datum_value>());
              WeylWord ww;
              lietype::Layout lo = check_involution(M,rd->val,ww);
              if (l==expression_base::no_value)
                return;
            
              for (unsigned int i=ww.size(); i-->0;)
                rd->val.simple_reflect(ww[i],M);
              std::unique_ptr<ComplexReductiveGroup>
                G(new ComplexReductiveGroup(rd->val,M));
              push_value(std::make_shared<inner_class_value>(std::move(G),lo));
              push_value(std::make_shared<vector_value>(std::vector<int>(ww.begin(),ww.end())));
              if (l==expression_base::single_value)
                wrap_tuple<2>();
            }
            
            void set_type_wrapper(expression_base::level l)
            { shared_string ict = get<string_value>();
                // and leave generators |gen| and type |lt|
              shared_value lt = *(execution_stack.end()-2);
              const LieType& type=force<Lie_type_value>(lt.get())->val;
              lietype::Layout lo(type,transform_inner_class_type(ict->val.c_str(),type));
            
              quotient_basis_wrapper(expression_base::single_value); 
              shared_value basis = pop_value();
            
              push_value(lt); push_value(basis);
              root_datum_wrapper(expression_base::single_value);
              shared_root_datum rd = get<root_datum_value>();
            
              push_value(lt); push_value(basis);
              push_value(ict);
              based_involution_wrapper(expression_base::single_value);
              shared_matrix M = get<matrix_value>();
              if (l==expression_base::no_value)
                return; // bow out now all possible errors are passed
            
              std::unique_ptr<ComplexReductiveGroup>
                G(new ComplexReductiveGroup(rd->val,M->val));
              push_value(std::make_shared<inner_class_value>(std::move(G),lo));
            }
            
            void set_inner_class_wrapper(expression_base::level l)
            { shared_string ict(get<string_value>());
              shared_root_datum rdv(get<root_datum_value>());
              const RootDatum& rd=rdv->val;
            
              lietype::Layout lo;
              lo.d_type = dynkin::Lie_type(rd.cartanMatrix(),true,false,lo.d_perm);
               // get (permuted) type
              for (size_t i=rd.semisimpleRank(); i<rd.rank(); ++i) // if not semisimple
              { lo.d_type.push_back(SimpleLieType('T',1)); // add a torus factor
                lo.d_perm.push_back(i);
                  // and a fixed point of permutation, needed by |lietype::involution|
              }
              lo.d_inner=transform_inner_class_type(ict->val.c_str(),lo.d_type);
            
              push_value(rdv);
              coroot_radical_wrapper(expression_base::single_value);
              transpose_mat_wrapper(expression_base::single_value);
              shared_matrix M(get<matrix_value>());
              size_t r=lo.d_type.rank();
              assert(M->val.numRows()==r and M->val.numRows()==r);
              ndebug_use(r);
            
              push_value(rdv);
              try
              { push_value(std::make_shared<matrix_value> 
                   (lietype::involution(lo).on_basis(M->val.columns())));
              }
              catch (std::runtime_error&) // relabel inexact division error
              { throw std::runtime_error 
                ("Inner class is not compatible with root datum");
            
              }
              fix_involution_wrapper(l);
            }
            
            void inner_class_equality_wrapper(expression_base::level l)
            { shared_inner_class G = get<inner_class_value>();
              shared_inner_class H = get<inner_class_value>();
              if (l!=expression_base::no_value)
                push_value(std::make_shared<bool_value>(&G->val==&H->val));
            }
            
            void distinguished_involution_wrapper(expression_base::level l)
            { shared_inner_class G = get<inner_class_value>();
              if (l!=expression_base::no_value)
                push_value(std::make_shared<matrix_value>(G->val.distinguished()));
            }
            
            void root_datum_of_inner_class_wrapper(expression_base::level l)
            { shared_inner_class G = get<inner_class_value>();
              if (l!=expression_base::no_value)
                push_value(std::make_shared<root_datum_value>(G->val.rootDatum()));
            }
            
            void inner_class_to_root_datum_coercion()
            { root_datum_of_inner_class_wrapper(expression_base::single_value); }
            
            void dual_inner_class_wrapper(expression_base::level l)
            { shared_inner_class G = get<inner_class_value>();
              if (l!=expression_base::no_value)
                push_value(std::make_shared<inner_class_value>(*G,tags::DualTag()));
            }
            
            void push_name_list(const realform_io::Interface& interface)
            { own_row result = std::make_shared<row_value>(0);
              for (size_t i=0; i<interface.numRealForms(); ++i)
                result->val.emplace_back
                  (std::make_shared<string_value>(interface.typeName(i)));
              push_value(std::move(result));
            }
            
            void form_names_wrapper(expression_base::level l)
            { shared_inner_class G = get<inner_class_value>();
              if (l!=expression_base::no_value)
                push_name_list(G->interface);
            }
            
            void dual_form_names_wrapper(expression_base::level l)
            { shared_inner_class G = get<inner_class_value>();
              if (l!=expression_base::no_value)
                push_name_list(G->dual_interface);
            }
            
            void n_real_forms_wrapper(expression_base::level l)
            { shared_inner_class G = get<inner_class_value>();
              if (l!=expression_base::no_value)
                push_value(std::make_shared<int_value>(G->val.numRealForms()));
            }
            
            void n_dual_real_forms_wrapper(expression_base::level l)
            { shared_inner_class G = get<inner_class_value>();
              if (l!=expression_base::no_value)
                push_value(std::make_shared<int_value>(G->val.numDualRealForms()));
            }
            
            void n_Cartan_classes_wrapper(expression_base::level l)
            { shared_inner_class G = get<inner_class_value>();
              if (l!=expression_base::no_value)
                push_value(std::make_shared<int_value>(G->val.numCartanClasses()));
            }
            
            void block_sizes_wrapper(expression_base::level l)
            { shared_inner_class G = get<inner_class_value>();
              if (l==expression_base::no_value)
                return;
              own_matrix M = std::make_shared<matrix_value> 
                (int_Matrix(G->val.numRealForms(),G->val.numDualRealForms()));
              for (size_t i = 0; i < M->val.numRows(); ++i)
                for (size_t j = 0; j < M->val.numColumns(); ++j)
                  M->val(i,j) =
                  G->val.block_size(G->interface.in(i),G->dual_interface.in(j));
              push_value(std::move(M));
            }
            
            void occurrence_matrix_wrapper(expression_base::level l)
            { shared_inner_class G = get<inner_class_value>();
              if (l==expression_base::no_value)
                return;
              size_t nr=G->val.numRealForms();
              size_t nc=G->val.numCartanClasses();
              own_matrix M = std::make_shared<matrix_value>(int_Matrix(nr,nc));
              for (size_t i=0; i<nr; ++i)
              { BitMap b=G->val.Cartan_set(G->interface.in(i));
                for (size_t j=0; j<nc; ++j)
                  M->val(i,j)= b.isMember(j) ? 1 : 0;
              }
              push_value(std::move(M));
            }
            
            void dual_occurrence_matrix_wrapper(expression_base::level l)
            { shared_inner_class G = get<inner_class_value>();
              if (l==expression_base::no_value)
                return;
              size_t nr=G->val.numDualRealForms();
              size_t nc=G->val.numCartanClasses();
              own_matrix M = std::make_shared<matrix_value>(int_Matrix(nr,nc));
              for (size_t i=0; i<nr; ++i)
              { BitMap b=G->val.dual_Cartan_set(G->dual_interface.in(i));
                for (size_t j=0; j<nc; ++j)
                  M->val(i,j)= b.isMember(j) ? 1 : 0;
              }
              push_value(std::move(M));
            }
            
            void real_form_wrapper(expression_base::level l)
            { shared_int i(get<int_value>());
              shared_inner_class G = get<inner_class_value>();
              if (size_t(i->val)>=G->val.numRealForms())
                throw std::runtime_error ("Illegal real form number: "+str(i->val));
            
              if (l!=expression_base::no_value)
                push_value(std::make_shared<real_form_value>(*G,G->interface.in(i->val)));
            }
            
            void form_number_wrapper(expression_base::level l)
            { shared_real_form rf= get<real_form_value>();
              if (l!=expression_base::no_value)
                push_value(std::make_shared<int_value>(rf->parent.interface.out(rf->val.realForm())));
            }
            
            void quasisplit_form_wrapper(expression_base::level l)
            { shared_inner_class G = get<inner_class_value>();
              if (l!=expression_base::no_value)
                push_value(std::make_shared<real_form_value>(*G,G->val.quasisplit()));
            }
            
            void inner_class_of_real_form_wrapper(expression_base::level l)
            { shared_real_form rf= get<real_form_value>();
              if (l!=expression_base::no_value)
                push_value(std::make_shared<inner_class_value>(rf->parent));
            }
            
            void real_form_to_inner_class_coercion()
            { inner_class_of_real_form_wrapper(expression_base::single_value); }
            
            void real_form_to_root_datum_coercion()
            { shared_real_form rf= get<real_form_value>();
              push_value(std::make_shared<root_datum_value>(rf->parent.val.rootDatum()));
            }
            
            void components_rank_wrapper(expression_base::level l)
            { shared_real_form R= get<real_form_value>();
              if (l!=expression_base::no_value)
                push_value(std::make_shared<int_value>(R->val.dualComponentReps().size()));
            }
            
            void count_Cartans_wrapper(expression_base::level l)
            { shared_real_form rf= get<real_form_value>();
              if (l!=expression_base::no_value)
                push_value(std::make_shared<int_value>(rf->val.numCartan()));
            }
            
            void KGB_size_wrapper(expression_base::level l)
            { shared_real_form rf= get<real_form_value>();
              if (l!=expression_base::no_value)
                push_value(std::make_shared<int_value>(rf->val.KGB_size()));
            }
            
            void base_grading_vector_wrapper(expression_base::level l)
            { own_real_form rf= non_const_get<real_form_value>();
              if (l!=expression_base::no_value)
              { const KGB& kgb=rf->kgb();
                RatCoweight t = kgb.base_grading_vector();
                push_value(std::make_shared<rational_vector_value>(t));
              }
            }
            
            void Cartan_order_matrix_wrapper(expression_base::level l)
            { shared_real_form rf= get<real_form_value>();
              if (l==expression_base::no_value)
                return;
              size_t n=rf->val.numCartan();
              own_matrix M = std::make_shared<matrix_value>(int_Matrix(n,n,0));
              const poset::Poset& p = rf->val.complexGroup().Cartan_ordering();
              for (size_t i=0; i<n; ++i)
                for (size_t j=i; j<n; ++j)
                  if (p.lesseq(i,j)) M->val(i,j)=1;
            
              push_value(std::move(M));
            }
            
            inline bool operator==
              (const RealReductiveGroup& x, const RealReductiveGroup& y)
            {
              return &x.complexGroup() == &y.complexGroup() and x.realForm()==y.realForm();
            }
            inline bool operator!=
              (const RealReductiveGroup& x, const RealReductiveGroup& y)
            { return not(x==y); }
            
            void real_form_equals_wrapper(expression_base::level l)
            { shared_real_form y = get<real_form_value>();
              shared_real_form x = get<real_form_value>();
              if (l==expression_base::no_value)
                return;
              push_value(std::make_shared<bool_value>(x->val==y->val));
            }
            
            void dual_real_form_wrapper(expression_base::level l)
            { shared_int i(get<int_value>());
              shared_inner_class G = get<inner_class_value>();
              if (size_t(i->val)>=G->val.numDualRealForms())
                throw std::runtime_error ("Illegal dual real form number: "+str(i->val));
            
              if (l==expression_base::no_value)
                return;
              inner_class_value G_check(*G,tags::DualTag());
               // tailor make an |inner_class_value|
              push_value(std::make_shared<real_form_value>(G_check ,G->dual_interface.in(i->val)));
            }
            
            void dual_quasisplit_form_wrapper(expression_base::level l)
            { shared_inner_class G = get<inner_class_value>();
              if (l==expression_base::no_value)
                return;
              inner_class_value G_check(*G,tags::DualTag());
               // tailor make an |inner_class_value|
              push_value(std::make_shared<real_form_value>(G_check,G->dual.quasisplit()));
            }
            
            void ic_Cartan_class_wrapper(expression_base::level l)
            { shared_int i(get<int_value>());
              shared_inner_class ic = get<inner_class_value>();
              if (size_t(i->val)>=ic->val.numCartanClasses())
                throw std::runtime_error ("Illegal Cartan class number: "+str(i->val)
            
                +", this inner class only has "+str(ic->val.numCartanClasses())
                +" of them");
              if (l!=expression_base::no_value)
                push_value(std::make_shared<Cartan_class_value>(*ic,i->val));
            }
            
            void rf_Cartan_class_wrapper(expression_base::level l)
            { shared_int i(get<int_value>());
              shared_real_form rf= get<real_form_value>();
              if (size_t(i->val)>=rf->val.numCartan())
                throw std::runtime_error ("Illegal Cartan class number: "+str(i->val)
            
                +", this real form only has "+str(rf->val.numCartan())+" of them");
              BitMap cs=rf->val.Cartan_set();
              if (l!=expression_base::no_value)
                push_value(std::make_shared<Cartan_class_value>(rf->parent,cs.n_th(i->val)));
            }
            
            void most_split_Cartan_wrapper(expression_base::level l)
            { shared_real_form rf= get<real_form_value>();
              if (l!=expression_base::no_value)
                push_value(std::make_shared<Cartan_class_value>(rf->parent,rf->val.mostSplit()));
            }
            
            TwistedInvolution twisted_from_involution
              (const ComplexReductiveGroup& G, const WeightInvolution theta)
            { const RootDatum& rd = G.rootDatum();
              unsigned int ssr =  rd.semisimpleRank();
              RootNbrList Delta(ssr);
              for (weyl::Generator i=0; i<ssr; ++i)
              { Delta[i]=rd.rootNbr(theta*rd.simpleRoot(i));
                if (Delta[i]==rd.numRoots()) // then image not found
                  throw std::runtime_error
                    ("Matrix maps simple root "+str(i)+" to non-root");
              }
              WeylWord ww = wrt_distinguished(rd,Delta);
              for (weyl::Generator i=0; i<ssr; ++i)
                if (Delta[i]!=G.twisted_root(rd.simpleRootNbr(i)))
                    // |Delta| should match distinguished involution
                  throw std::runtime_error
                    ("Matrix does not define a root datum automorphism");
              return G.weylGroup().element(ww);
            }
            
            void Cartan_class_real_form_wrapper(expression_base::level l)
            { shared_rational_vector grading_shift = get<rational_vector_value>();
              shared_matrix theta = get<matrix_value>();
              shared_inner_class G = get<inner_class_value>();
              { Coweight num(grading_shift->val.numerator().begin(),
                             grading_shift->val.numerator().end());
                if (theta->val.right_prod(num)!=num)
                  throw std::runtime_error ("Grading coweight not fixed by involution");
            
              }
            
              TwistedInvolution tw = twisted_from_involution(G->val,theta->val);
            
              CartanNbr cn; WeylWord ww; cartanclass::AdjointFiberElt rep;
              RealFormNbr rf = real_form_of(G->val,tw,grading_shift->val,cn,ww,rep);
              if (l==expression_base::no_value)
                 return;
              push_value(std::make_shared<Cartan_class_value>(*G,cn));
              push_value(std::make_shared<real_form_value>(*G,rf));
              if (l==expression_base::single_value)
                wrap_tuple<2>();
            }
            
            void Cartan_involution_wrapper(expression_base::level l)
            { shared_Cartan_class cc(get<Cartan_class_value>());
              if (l==expression_base::no_value)
                return;
              push_value(std::make_shared<matrix_value>(cc->val.involution()));
            }
            
            void Cartan_info_wrapper(expression_base::level l)
            { shared_Cartan_class cc(get<Cartan_class_value>());
              if (l==expression_base::no_value)
                return;
            
              push_value(std::make_shared<int_value>(cc->val.fiber().torus().compactRank()));
              push_value(std::make_shared<int_value>(cc->val.fiber().torus().complexRank()));
              push_value(std::make_shared<int_value>(cc->val.fiber().torus().splitRank()));
              wrap_tuple<3>();
            
              const weyl::TwistedInvolution& tw =
                cc->parent.val.involution_of_Cartan(cc->number);
              WeylWord ww = cc->parent.val.weylGroup().word(tw);
            
              std::vector<int> v(ww.begin(),ww.end());
              push_value(std::make_shared<vector_value>(v));
            
              push_value(std::make_shared<int_value>(cc->val.orbitSize()));
              push_value(std::make_shared<int_value>(cc->val.fiber().fiberSize()));
              wrap_tuple<2>();
            
              const RootSystem& rs=cc->parent.val.rootDatum();
            
            // print types of imaginary and real root systems and of Complex factor
              push_value(std::make_shared<Lie_type_value>(rs.Lie_type(cc->val.simpleImaginary())));
              push_value(std::make_shared<Lie_type_value>(rs.Lie_type(cc->val.simpleReal())));
              push_value(std::make_shared<Lie_type_value>(rs.Lie_type(cc->val.simpleComplex())));
              wrap_tuple<3>();
            
              if (l==expression_base::single_value)
                wrap_tuple<4>();
            }
            
            void real_forms_of_Cartan_wrapper(expression_base::level l)
            { shared_Cartan_class cc(get<Cartan_class_value>());
              if (l==expression_base::no_value)
                return;
              const inner_class_value& ic=cc->parent;
              own_row result = std::make_shared<row_value>(cc->val.numRealForms());
              for (size_t i=0,k=0; i<ic.val.numRealForms(); ++i)
              { RealFormNbr rf = ic.interface.in(i);
                BitMap b(ic.val.Cartan_set(rf));
                if (b.isMember(cc->number))
                  result->val[k++] = std::make_shared<real_form_value>(ic,rf);
              }
              push_value(std::move(result));
            }
            
            void dual_real_forms_of_Cartan_wrapper(expression_base::level l)
            { shared_Cartan_class cc(get<Cartan_class_value>());
              if (l==expression_base::no_value)
                return;
              const inner_class_value dual_ic(cc->parent,tags::DualTag());
              own_row result = std::make_shared<row_value>(cc->val.numDualRealForms());
              for (size_t i=0,k=0; i<dual_ic.val.numRealForms(); ++i)
              { RealFormNbr drf = cc->parent.dual_interface.in(i);
                BitMap b (cc->parent.val.dual_Cartan_set(drf));
                if (b.isMember(cc->number))
                  result->val[k++] = std::make_shared<real_form_value>(dual_ic,drf);
              }
              push_value(std::move(result));
            }
            
            void fiber_partition_wrapper(expression_base::level l)
            { shared_real_form rf= get<real_form_value>();
              shared_Cartan_class cc(get<Cartan_class_value>());
              if (&rf->parent.val!=&cc->parent.val)
                throw std::runtime_error
                ("Inner class mismatch between real form and Cartan class");
            
              BitMap b(cc->parent.val.Cartan_set(rf->val.realForm()));
              if (!b.isMember(cc->number))
                throw std::runtime_error
                ("Cartan class not defined for this real form");
            
              if (l==expression_base::no_value)
                return;
            
              const Partition& pi = cc->val.fiber().weakReal();
              const RealFormNbrList rf_nr=
                 cc->parent.val.realFormLabels(cc->number);
                 // translate part number of |pi| to real form
              own_row result =
                std::make_shared<row_value>(0); // cannot predict exact size here
              for (size_t i=0; i<pi.size(); ++i)
                if (rf_nr[pi.class_of(i)] == rf->val.realForm())
                  result->val.push_back(std::make_shared<int_value>(i));
              push_value(std::move(result));
            }
            
            void square_classes_wrapper(expression_base::level l)
            { shared_Cartan_class cc(get<Cartan_class_value>());
              const realform_io::Interface rfi = cc->parent.interface;
              const RealFormNbrList& rfl = cc->parent.val.realFormLabels(cc->number);
              if (l==expression_base::no_value)
                return;
              size_t n_sq_classes = cc->val.numRealFormClasses();
              own_row result = std::make_shared<row_value>(n_sq_classes);
              for (cartanclass::square_class csc=0; csc<n_sq_classes; ++csc)
              { const Partition& pi = cc->val.fiber_partition(csc);
                own_row part = std::make_shared<row_value>(pi.classCount());
                for (unsigned long c=0; c<pi.classCount(); ++c)
                   part->val[c] =
                      std::make_shared<int_value>(rfi.out(rfl[cc->val.toWeakReal(c,csc)]));
                result->val[csc] = std::move(part);
              }
              push_value(std::move(result));
            }
            
            void print_gradings_wrapper(expression_base::level l)
            { shared_real_form rf= get<real_form_value>();
            shared_Cartan_class cc(get<Cartan_class_value>());
              if (&rf->parent.val!=&cc->parent.val)
                throw std::runtime_error
                ("Inner class mismatch between real form and Cartan class");
            
              BitMap b(cc->parent.val.Cartan_set(rf->val.realForm()));
              if (!b.isMember(cc->number))
                throw std::runtime_error ("Cartan class not defined for this real form");
            
            
              const Partition& pi = cc->val.fiber().weakReal();
              const RealFormNbrList rf_nr=
                 cc->parent.val.realFormLabels(cc->number);
                 // translate part number of |pi| to real form
            
              const RootNbrList& si = cc->val.fiber().simpleImaginary();
              // simple imaginary roots
            
              int_Matrix cm;
              Permutation sigma;
            
            { cm=cc->parent.val.rootDatum().cartanMatrix(si);
              dynkin::DynkinDiagram d(cm); sigma = dynkin::bourbaki(d);
            }
              
              { *output_stream << "Imaginary root system is ";
                if (si.size()==0) *output_stream<<"empty.\n";
                else
                { LieType t = dynkin::Lie_type(cm);
              
                  *output_stream << "of type " << t << ", with simple root"
                            << (si.size()==1 ? " " : "s ");
                  for (size_t i=0; i<si.size(); ++i)
                    *output_stream << si[sigma[i]] << (i<si.size()-1 ? "," : ".\n");
                }
              }
              
              { bool first=true; std::ostringstream os;
                for (size_t i=0; i<pi.size(); ++i)
                  if ( rf_nr[pi.class_of(i)] == rf->val.realForm())
                  { os << ( first ? first=false,'[' : ',');
                    Grading gr=cc->val.fiber().grading(i);
                    gr=sigma.pull_back(gr);
                    prettyprint::prettyPrint(os,gr,si.size());
                  }
                os << "]" << std::endl;
                ioutils::foldLine(*output_stream,os.str(),"",",");
              }
              if (l==expression_base::single_value)
                wrap_tuple<0>();
            }
            
            void KGB_elt_wrapper(expression_base::level l)
            { int i = get<int_value>()->val;
              own_real_form rf= non_const_get<real_form_value>();
              if (size_t(i)>=rf->val.KGB_size())
                throw std::runtime_error ("Inexistent KGB element: "+str(i));
            
              if (l!=expression_base::no_value)
                push_value(std::make_shared<KGB_elt_value>(rf,i));
            }
            
            void decompose_KGB_wrapper(expression_base::level l)
            { shared_KGB_elt x = get<KGB_elt_value>();
              if (l==expression_base::no_value)
                return;
              push_value(x->rf);
              push_value(std::make_shared<int_value>(x->val));
              if (l==expression_base::single_value)
                wrap_tuple<2>();
            }
            
            void KGB_Cartan_wrapper(expression_base::level l)
            { shared_KGB_elt x = get<KGB_elt_value>();
              const KGB& kgb=x->rf->kgb();
              if (l!=expression_base::no_value)
                push_value(std::make_shared<Cartan_class_value>(x->rf->parent,kgb.Cartan_class(x->val)));
            }
            
            void KGB_involution_wrapper(expression_base::level l)
            { shared_KGB_elt x = get<KGB_elt_value>();
              const KGB& kgb=x->rf->kgb();
              const ComplexReductiveGroup& G=x->rf->val.complexGroup();
              if (l!=expression_base::no_value)
                push_value(std::make_shared<matrix_value>(G.involutionMatrix(kgb.involution(x->val))));
            }
            
            void KGB_length_wrapper(expression_base::level l)
            { shared_KGB_elt x = get<KGB_elt_value>();
              const KGB& kgb=x->rf->kgb();
              if (l!=expression_base::no_value)
                push_value(std::make_shared<int_value>(kgb.length(x->val)));
            }
            
            void KGB_cross_wrapper(expression_base::level l)
            { own_KGB_elt x = get_own<KGB_elt_value>();
              const KGB& kgb=x->rf->kgb();
              int s = get<int_value>()->val;
              if (static_cast<unsigned>(s)>=kgb.rank())
                throw std::runtime_error ("Illegal simple reflection: "+str(s));
              if (l==expression_base::no_value)
                return;
              x->val= kgb.cross(s,x->val); // do cross action
              push_value(x);
            }
            
            void KGB_Cayley_wrapper(expression_base::level l)
            { own_KGB_elt x = get_own<KGB_elt_value>();
              const KGB& kgb=x->rf->kgb();
              int s = get<int_value>()->val;
              if (static_cast<unsigned>(s)>=kgb.rank())
                throw std::runtime_error ("Illegal simple reflection: "+str(s));
              if (l==expression_base::no_value)
                return;
              if (kgb.cayley(s,x->val)!=UndefKGB) // when defined
                x->val= kgb.cayley(s,x->val); // do Cayley transform
              push_value(x);
            }
            
            void KGB_inv_Cayley_wrapper(expression_base::level l)
            { own_KGB_elt x = get_own<KGB_elt_value>();
              const KGB& kgb=x->rf->kgb();
              int s = get<int_value>()->val;
              if (static_cast<unsigned>(s)>=kgb.rank())
                throw std::runtime_error ("Illegal simple reflection: "+str(s));
              if (l==expression_base::no_value)
                return;
              if (kgb.inverseCayley(s,x->val).first!=UndefKGB) // when defined, do first
                x->val= kgb.inverseCayley(s,x->val).first; // inverse Cayley transform
              push_value(x);
            }
            
            void KGB_status_wrapper(expression_base::level l)
            { shared_KGB_elt x = get<KGB_elt_value>();
              const KGB& kgb=x->rf->kgb();
              int s = get<int_value>()->val;
              if (static_cast<unsigned>(s)>=kgb.rank())
                throw std::runtime_error ("Illegal simple reflection: "+str(s));
              if (l==expression_base::no_value)
                return;
              unsigned stat=kgb.status(s,x->val);
              push_value(std::make_shared<int_value>
                (stat==0 and not kgb.isDescent(s,x->val) ? 4 : stat));
            }
            
            void root_status_wrapper(expression_base::level l)
            { shared_KGB_elt x = get<KGB_elt_value>();
              const KGB& kgb=x->rf->kgb();
              const ComplexReductiveGroup& G=x->rf->parent.val;
              const RootDatum& rd = G.rootDatum();
              shared_vector alpha_vec = get<vector_value>();
              RootNbr alpha = rd.rootNbr(alpha_vec->val);
              if (alpha>=rd.numRoots())
                throw std::runtime_error ("Vector is not a root");
            
              if (l==expression_base::no_value)
                return;
              unsigned stat=kgb::status(kgb,x->val, rd,alpha);
              if (stat==0) // $\alpha$ is a complex root, check if it is an ascent
              {
                RootNbr theta_alpha =
                  G.involution_table().root_involution(kgb.inv_nr(x->val),alpha);
                if (rd.isPosRoot(alpha)==rd.isPosRoot(theta_alpha))
                  stat = 4; // set status to complex ascent
              }
              push_value(std::make_shared<int_value> (stat));
            }
            
            void build_KGB_element_wrapper(expression_base::level l)
            { shared_rational_vector grading_shift = get<rational_vector_value>();
              shared_matrix theta = get<matrix_value>();
              own_real_form rf = non_const_get<real_form_value>();
            
              { Coweight num(grading_shift->val.numerator().begin(),
                             grading_shift->val.numerator().end());
                if (theta->val.right_prod(num)!=num)
                  throw std::runtime_error ("Grading coweight not fixed by involution");
            
              }
              RatCoweight tv = grading_shift->val - rf->kgb().base_grading_vector();
              if (tv.normalize().denominator()!=1)
                throw std::runtime_error("Grading coweight not appropriate for real form");
            
              const ComplexReductiveGroup& G = rf->parent.val;
              TorusPart t(tv.numerator());
              TwistedInvolution tw = twisted_from_involution(G,theta->val);
              TitsElt a (G.titsGroup(),t,tw);
            
              KGBElt x = rf->kgb().lookup(a);
              if (x== rf->kgb().size())
                throw std::runtime_error("KGB element not present");
            
              if (l==expression_base::no_value)
                return;
              push_value(std::make_shared<KGB_elt_value> (rf,x));
            }
            
            void KGB_twist_wrapper(expression_base::level l)
            { own_KGB_elt x = get_own<KGB_elt_value>();
              const KGB& kgb=x->rf->kgb();
              if (l==expression_base::no_value)
                return;
              x->val= kgb.Hermitian_dual(x->val); // do twist
              push_value(x);
            }
            
            void torus_bits_wrapper(expression_base::level l)
            { shared_KGB_elt x = get<KGB_elt_value>();
              if (l!=expression_base::no_value)
              { const KGB& kgb=x->rf->kgb();
                TorusPart t = kgb.torus_part(x->val);
                own_vector result = std::make_shared<vector_value>
                  (int_Vector(kgb.torus_rank(),0));
                int_Vector& v = result->val;
                for (unsigned int i=0; i<kgb.torus_rank(); ++i)
                  v[i]=t[i]; // lift from binary to integer
                push_value(std::move(result));
              }
            }
            
            void torus_factor_wrapper(expression_base::level l)
            { shared_KGB_elt x = get<KGB_elt_value>();
              if (l!=expression_base::no_value)
              { const KGB& kgb=x->rf->kgb();
                RatCoweight t = kgb.torus_part_global(x->val);
                push_value(std::make_shared<rational_vector_value>(t));
              }
            }
            
            void KGB_equals_wrapper(expression_base::level l)
            { shared_KGB_elt y = get<KGB_elt_value>();
              shared_KGB_elt x = get<KGB_elt_value>();
              if (l==expression_base::no_value)
                return;
              push_value(std::make_shared<bool_value>(x->rf->val==y->rf->val and x->val==y->val));
            }
            
            void Fokko_block_wrapper(expression_base::level l)
            { own_real_form drf=non_const_get<real_form_value>();
              own_real_form rf=non_const_get<real_form_value>();
            
              if (&rf->parent.dual!=&drf->parent.val)
                throw std::runtime_error 
                ("Inner class mismatch between real form and dual real form");
            
              BitMap b(rf->parent.val.dual_Cartan_set(drf->val.realForm()));
              if (not b.isMember(rf->val.mostSplit()))
                throw std::runtime_error 
                ("Real form and dual real form are incompatible");
            
              if (l!=expression_base::no_value)
                push_value(std::make_shared<Block_value>(rf,drf));
            }
            
            void decompose_block_wrapper(expression_base::level l)
            { shared_Block b = get<Block_value>();
              if (l==expression_base::no_value)
                return;
              push_value(b->rf);
              push_value(b->dual_rf);
              if (l==expression_base::single_value)
                wrap_tuple<2>();
            }
            
            void block_size_wrapper(expression_base::level l)
            { shared_Block b = get<Block_value>();
              if (l!=expression_base::no_value)
                push_value(std::make_shared<int_value>(b->val.size()));
            }
            
            void block_element_wrapper(expression_base::level l)
            { shared_int i(get<int_value>());
              shared_Block b = get<Block_value>();
              BlockElt z = i->val; // extract value unsigned
              if (z>=b->val.size())
                throw std::runtime_error
                  ("Block element " +str(z) + " out of range (<" + str(b->val.size())+")");
              if (l==expression_base::no_value)
                return;
              push_value(std::make_shared<KGB_elt_value>(b->rf,b->val.x(z)));
              inner_class_value dual_ic(b->rf->parent,tags::DualTag());
              own_real_form drf =
                std::make_shared<real_form_value>(dual_ic,b->dual_rf->val.realForm());
              push_value(std::make_shared<KGB_elt_value>(drf,b->val.y(z)));
              if (l==expression_base::single_value)
                wrap_tuple<2>();
            }
            
            void block_index_wrapper(expression_base::level l)
            { shared_KGB_elt y = get<KGB_elt_value>();
              shared_KGB_elt x = get<KGB_elt_value>();
              shared_Block b = get<Block_value>();
              if (&b->rf->parent.val!=&x->rf->parent.val)
                throw std::runtime_error ("Real form not in inner class of block");
              if (&b->rf->parent.val!=&y->rf->parent.dual)
                throw std::runtime_error ("Dual real form not in inner class of block");
              const KGB& kgb = b->rf->kgb(); const KGB& dual_kgb = b->dual_rf->kgb();
              const TwistedWeylGroup& tw = kgb.twistedWeylGroup();
              const TwistedWeylGroup& dual_tw = dual_kgb.twistedWeylGroup();
              if (blocks::dual_involution(kgb.involution(x->val),tw,dual_tw)
                  != dual_kgb.involution(y->val))
                throw std::runtime_error ("Fiber mismatch KGB and dual KGB elements");
            
              BlockElt z = b->val.element(x->val,y->val);
            
              if (l==expression_base::no_value)
                return;
              push_value(std::make_shared<int_value>(z));
            }
            
            void dual_block_wrapper(expression_base::level l)
            { shared_Block b = get<Block_value>();
              if (l!=expression_base::no_value)
                push_value(std::make_shared<Block_value>(b->dual_rf,b->rf));
            }
            
            void block_status_wrapper(expression_base::level l)
            { BlockElt i = get<int_value>()->val;
              shared_Block b = get<Block_value>();
              unsigned int s = get<int_value>()->val;
              if (s>=b->rf->val.semisimpleRank())
                throw std::runtime_error ("Illegal simple reflection: "+str(s));
              if (i>=b->val.size())
                throw std::runtime_error
                  ("Block element " +str(i) + " out of range (<" + str(b->val.size())+")");
              if (l==expression_base::no_value)
                return;
            
              const DescentStatus::Value dv = b->val.descentValue(s,i);
            // renumber from |DescentStatus::Value| order to C-,ic,r1,r2,C+,rn,i1,i2
              static const unsigned char tab [] = {4,5,6,7,1,0,3,2};
              push_value(std::make_shared<int_value>(tab[dv]));
            }
            
            void block_cross_wrapper(expression_base::level l)
            { BlockElt i = get<int_value>()->val;
              shared_Block b = get<Block_value>();
              unsigned int s = get<int_value>()->val;
              if (s>=b->rf->val.semisimpleRank())
                throw std::runtime_error ("Illegal simple reflection: "+str(s));
              if (i>=b->val.size())
                throw std::runtime_error
                  ("Block element " +str(i) + " out of range (<" + str(b->val.size())+")");
              if (l==expression_base::no_value)
                return;
              push_value(std::make_shared<int_value>(b->val.cross(s,i)));
            }
            
            void block_Cayley_wrapper(expression_base::level l)
            { shared_int i = get<int_value>();
              shared_Block b = get<Block_value>();
              unsigned int s = get<int_value>()->val;
              if (s>=b->rf->val.semisimpleRank())
                throw std::runtime_error ("Illegal simple reflection: "+str(s));
              if (static_cast<unsigned int>(i->val) >= b->val.size())
                throw std::runtime_error
                  ("Block element " +str(i) + " out of range (<" + str(b->val.size())+")");
              if (l==expression_base::no_value)
                return;
              BlockElt sx = b->val.cayley(s,i->val).first;
              if (sx==UndefBlock) // when undefined, return i to indicate so
                push_value(i);
              else
                push_value(std::make_shared<int_value>(sx));
            }
            
            void block_inverse_Cayley_wrapper(expression_base::level l)
            { shared_int i = get<int_value>();
              shared_Block b = get<Block_value>();
              unsigned int s = get<int_value>()->val;
              if (s>=b->rf->val.semisimpleRank())
                throw std::runtime_error ("Illegal simple reflection: "+str(s));
              if (static_cast<unsigned int>(i->val) >= b->val.size())
                throw std::runtime_error
                  ("Block element " +str(i) + " out of range (<" + str(b->val.size())+")");
              if (l==expression_base::no_value)
                return;
              BlockElt sx = b->val.inverseCayley(s,i->val).first;
              if (sx==UndefBlock) // when undefined, return i to indicate so
                push_value(i);
              else
                push_value(std::make_shared<int_value>(sx));
            }
            
            void module_parameter_wrapper(expression_base::level l)
            { shared_rational_vector nu(get<rational_vector_value>());
              shared_vector lam_rho(get<vector_value>());
              shared_KGB_elt x = get<KGB_elt_value>();
              if (nu->val.size()!=lam_rho->val.size()
                  or nu->val.size()!=x->rf->val.rank())
                throw std::runtime_error ("Rank mismatch: ("
                    +str(x->rf->val.rank())+","
            	+str(lam_rho->val.size())+","+str(nu->val.size())+")");
            
              if (l!=expression_base::no_value)
                push_value(std::make_shared<module_parameter_value>  (x->rf,
                  x->rf->rc().sr(x->val,lam_rho->val,nu->val)));
            }
            
            void unwrap_parameter_wrapper(expression_base::level l)
            { shared_module_parameter p = get<module_parameter_value>();
              if (l!=expression_base::no_value)
              { push_value(std::make_shared<KGB_elt_value>(p->rf,p->val.x()));
                push_value(std::make_shared<vector_value>(p->rc().lambda_rho(p->val)));
                push_value(std::make_shared<rational_vector_value>(p->rc().nu(p->val)));
                if (l==expression_base::single_value)
                  wrap_tuple<3>();
              }
            }
            
            void infinitesimal_character_wrapper(expression_base::level l)
            { shared_module_parameter p = get<module_parameter_value>();
              if (l!=expression_base::no_value)
                push_value(std::make_shared<rational_vector_value>(p->val.gamma()));
            }
            
            void real_form_of_parameter_wrapper(expression_base::level l)
            { shared_module_parameter p = get<module_parameter_value>();
              if (l!=expression_base::no_value)
                push_value(p->rf);
            }
            
            void is_standard_wrapper(expression_base::level l)
            { shared_module_parameter p = get<module_parameter_value>();
              RootNbr witness;
              if (l!=expression_base::no_value)
                push_value(std::make_shared<bool_value>(p->rc().is_standard(p->val,witness)));
            }
            
            void is_zero_wrapper(expression_base::level l)
            { shared_module_parameter p = get<module_parameter_value>();
              RootNbr witness;
              if (l!=expression_base::no_value)
                push_value(std::make_shared<bool_value>(p->rc().is_zero(p->val,witness)));
            }
            
            void is_final_wrapper(expression_base::level l)
            { shared_module_parameter p = get<module_parameter_value>();
              RootNbr witness;
              if (l!=expression_base::no_value)
                push_value(std::make_shared<bool_value>(p->rc().is_final(p->val,witness)));
            }
            
            void parameter_dominant_wrapper(expression_base::level l)
            { own_module_parameter p = get_own<module_parameter_value>();
              if (l!=expression_base::no_value)
              { p->rc().make_dominant(p->val);
                push_value(p);
              }
            }
            
            void parameter_equivalent_wrapper(expression_base::level l)
            { shared_module_parameter q = get<module_parameter_value>();
              shared_module_parameter p = get<module_parameter_value>();
              if (p->rf->val!=q->rf->val)
                throw std::runtime_error 
                  ("Real form mismatch when testing equivalence");
              if (l!=expression_base::no_value)
              { StandardRepr z0=p->val, z1=q->val; // copy
                p->rc().make_dominant(z0); q->rc().make_dominant(z1);
                push_value(std::make_shared<bool_value>(z0==z1));
              }
            }
            
            void parameter_cross_wrapper(expression_base::level l)
            { shared_module_parameter p = get<module_parameter_value>();
              int s = get<int_value>()->val;
              unsigned int r =
                rootdata::integrality_rank(p->rf->val.rootDatum(),p->val.gamma());
              if (static_cast<unsigned>(s)>=r)
                throw std::runtime_error
                  ("Illegal simple reflection: "+str(s)+ ", should be <"+str(r));
              if (l!=expression_base::no_value)
                push_value(std::make_shared<module_parameter_value>(p->rf,p->rc().cross(s,p->val)));
            }
            
            void parameter_Cayley_wrapper(expression_base::level l)
            { shared_module_parameter p = get<module_parameter_value>();
              int s = get<int_value>()->val;
              unsigned int r =
                rootdata::integrality_rank(p->rf->val.rootDatum(),p->val.gamma());
              if (static_cast<unsigned>(s)>=r)
                throw std::runtime_error
                  ("Illegal simple reflection: "+str(s)+ ", should be <"+str(r));
              if (l!=expression_base::no_value)
                push_value(std::make_shared<module_parameter_value>(p->rf,p->rc().Cayley(s,p->val)));
            }
            
            void parameter_inv_Cayley_wrapper(expression_base::level l)
            { shared_module_parameter p = get<module_parameter_value>();
              int s = get<int_value>()->val;
              unsigned int r =
                rootdata::integrality_rank(p->rf->val.rootDatum(),p->val.gamma());
              if (static_cast<unsigned>(s)>=r)
                throw std::runtime_error
                  ("Illegal simple reflection: "+str(s)+ ", should be <"+str(r));
              if (l!=expression_base::no_value)
                push_value(std::make_shared<module_parameter_value>(p->rf,p->rc().inv_Cayley(s,p->val)));
            }
            
            void parameter_twist_wrapper(expression_base::level l)
            { shared_module_parameter p = get<module_parameter_value>();
              if (l!=expression_base::no_value)
                push_value(std::make_shared<module_parameter_value>(p->rf,p->rc().twist(p->val)));
            }
            
            void orientation_number_wrapper(expression_base::level l)
            { shared_module_parameter p = get<module_parameter_value>();
              if (l!=expression_base::no_value)
                push_value(std::make_shared<int_value>(p->rc().orientation_number(p->val)));
            }
            
            void reducibility_points_wrapper(expression_base::level l)
            { shared_module_parameter p = get<module_parameter_value>();
              if (l!=expression_base::no_value)
              {
                RationalList rp = p->rc().reducibility_points(p->val);
                  // method normalises rationals
                own_row result = std::make_shared<row_value>(rp.size());
                for (size_t i=0; i<rp.size(); ++i)
                  result->val[i]=std::make_shared<rat_value>(rp[i]);
                push_value(std::move(result));
              }
            }
            
            void test_standard(const module_parameter_value& p)
            { RootNbr witness;
              if (p.rc().is_standard(p.val,witness))
                return;
              std::ostringstream os; p.print(os);
              os << "\nParameter not standard, negative on coroot #" << witness;
              throw std::runtime_error(os.str());
            }
            
            void test_nonzero_final(const module_parameter_value& p)
            { RootNbr witness; bool zero=p.rc().is_zero(p.val,witness);
              if (not zero and p.rc().is_final(p.val,witness))
                return; // nothing to report
              std::ostringstream os; p.print(os);
            os << "\nParameter is " << (zero ? "zero" : "not final")
                 <<", as witnessed by coroot #" << witness;
              throw std::runtime_error(os.str());
            }
            
            void print_n_block_wrapper(expression_base::level l)
            { shared_module_parameter p = get<module_parameter_value>();
              test_standard(*p);
              BlockElt init_index; // will hold index in the block of the initial element
              non_integral_block block(p->rc(),p->val,init_index);
              *output_stream << "Parameter defines element " << init_index
                           << " of the following block:" << std::endl;
              block.print_to(*output_stream,true);
                // print block using involution expressions
              if (l==expression_base::single_value)
                wrap_tuple<0>();
            }
            
            void block_wrapper(expression_base::level l)
            { shared_module_parameter p = get<module_parameter_value>();
              test_standard(*p);
              if (l!=expression_base::no_value)
              {
                BlockElt start; // will hold index in the block of the initial element
                non_integral_block block(p->rc(),p->val,start);
                
                { own_row param_list = std::make_shared<row_value>(block.size());
                  const RatWeight& gamma=block.gamma();
                  for (BlockElt z=0; z<block.size(); ++z)
                  { StandardRepr block_elt_param =
                      p->rc().sr(block.parent_x(z),block.lambda_rho(z),gamma);
                    param_list->val[z] =
                	std::make_shared<module_parameter_value>(p->rf,block_elt_param);
                  }
                  push_value(std::move(param_list));
                
                }
                push_value(std::make_shared<int_value>(start));
                if (l==expression_base::single_value)
                  wrap_tuple<2>();
              }
            }
            
            void partial_block_wrapper(expression_base::level l)
            { shared_module_parameter p = get<module_parameter_value>();
              test_standard(*p);
              if (l!=expression_base::no_value)
              {
                non_integral_block block(p->rc(),p->val);
                
                { own_row param_list = std::make_shared<row_value>(block.size());
                  const RatWeight& gamma=block.gamma();
                  for (BlockElt z=0; z<block.size(); ++z)
                  { StandardRepr block_elt_param =
                      p->rc().sr(block.parent_x(z),block.lambda_rho(z),gamma);
                    param_list->val[z] =
                	std::make_shared<module_parameter_value>(p->rf,block_elt_param);
                  }
                  push_value(std::move(param_list));
                
                }
              }
            }
            
            void param_length_wrapper(expression_base::level l)
            { shared_module_parameter p = get<module_parameter_value>();
              test_standard(*p);
              if (l!=expression_base::no_value)
                push_value(std::make_shared<int_value>(p->rt().length(p->val)));
            }
            
            void KL_block_wrapper(expression_base::level l)
            { shared_module_parameter p = get<module_parameter_value>();
              test_standard(*p);
              if (l!=expression_base::no_value)
              {
                BlockElt start; // will hold index in the block of the initial element
                non_integral_block block(p->rc(),p->val,start);
                
                { own_row param_list = std::make_shared<row_value>(block.size());
                  const RatWeight& gamma=block.gamma();
                  for (BlockElt z=0; z<block.size(); ++z)
                  { StandardRepr block_elt_param =
                      p->rc().sr(block.parent_x(z),block.lambda_rho(z),gamma);
                    param_list->val[z] =
                	std::make_shared<module_parameter_value>(p->rf,block_elt_param);
                  }
                  push_value(std::move(param_list));
                
                }
                push_value(std::make_shared<int_value>(start));
                const kl::KLContext& klc = block.klc(block.size()-1,false);
            
                own_matrix M = std::make_shared<matrix_value>(int_Matrix(klc.size()));
                for (size_t y=1; y<klc.size(); ++y)
                  for (size_t x=0; x<y; ++x)
                    M->val(x,y)= klc.KL_pol_index(x,y);
            
                own_row polys = std::make_shared<row_value>(0);
                polys->val.reserve(klc.polStore().size());
                for (size_t i=0; i<klc.polStore().size(); ++i)
                {
                  const kl::KLPol& pol = klc.polStore()[i];
                  std::vector<int> coeffs(pol.size());
                  for (size_t j=pol.size(); j-->0; )
                    coeffs[j]=pol[j];
                  polys->val.emplace_back(std::make_shared<vector_value>(coeffs));
                }
            
                own_vector length_stops = std::make_shared<vector_value>(
                   int_Vector(block.length(block.size()-1)+1));
                length_stops->val[0]=0;
                for (size_t i=1; i<length_stops->val.size(); ++i)
                  length_stops->val[i]=block.length_first(i);
            
                unsigned n_survivors=0;
                for (BlockElt z=0; z<block.size(); ++z)
                  if (block.survives(z))
                    ++n_survivors;
                own_vector survivor =
                  std::make_shared<vector_value>(int_Vector(n_survivors));
                { unsigned i=0;
                  for (BlockElt z=0; z<block.size(); ++z)
                    if (block.survives(z))
                      survivor->val[i++]=z;
                  assert(i==n_survivors);
                }
                own_matrix contributes_to = std::make_shared<matrix_value>(
                  int_Matrix(n_survivors,block.size(),0));
                for (BlockElt z=0; z<block.size(); ++z)
                { BlockEltList sb = block.survivors_below(z);
                  for (BlockEltList::const_iterator it=sb.begin(); it!=sb.end(); ++it)
                  { BlockElt x= permutations::find_index<int>(survivor->val,*it);
                      // a row index
                    if ((block.length(z)-block.length(*it))%2==0)
                      ++contributes_to->val(x,z);
                    else
                      --contributes_to->val(x,z);
                  }
                }
            
                push_value(std::move(M));
                push_value(std::move(polys));
                push_value(std::move(length_stops));
                push_value(std::move(survivor));
                push_value(std::move(contributes_to));
            
                if (l==expression_base::single_value)
                  wrap_tuple<7>();
              }
            }
            
            void partial_KL_block_wrapper(expression_base::level l)
            { shared_module_parameter p = get<module_parameter_value>();
              test_standard(*p);
              if (l!=expression_base::no_value)
              {
                non_integral_block block(p->rc(),p->val);
                
                { own_row param_list = std::make_shared<row_value>(block.size());
                  const RatWeight& gamma=block.gamma();
                  for (BlockElt z=0; z<block.size(); ++z)
                  { StandardRepr block_elt_param =
                      p->rc().sr(block.parent_x(z),block.lambda_rho(z),gamma);
                    param_list->val[z] =
                	std::make_shared<module_parameter_value>(p->rf,block_elt_param);
                  }
                  push_value(std::move(param_list));
                
                }
                const kl::KLContext& klc = block.klc(block.size()-1,false);
                // compute KL polynomials, silently
            
                own_matrix M = std::make_shared<matrix_value>(int_Matrix(klc.size()));
                for (size_t y=1; y<klc.size(); ++y)
                  for (size_t x=0; x<y; ++x)
                    M->val(x,y)= klc.KL_pol_index(x,y);
            
                own_row polys = std::make_shared<row_value>(0);
                polys->val.reserve(klc.polStore().size());
                for (size_t i=0; i<klc.polStore().size(); ++i)
                {
                  const kl::KLPol& pol = klc.polStore()[i];
                  std::vector<int> coeffs(pol.size());
                  for (size_t j=pol.size(); j-->0; )
                    coeffs[j]=pol[j];
                  polys->val.emplace_back(std::make_shared<vector_value>(coeffs));
                }
            
                own_vector length_stops = std::make_shared<vector_value>(
                   int_Vector(block.length(block.size()-1)+1));
                length_stops->val[0]=0;
                for (size_t i=1; i<length_stops->val.size(); ++i)
                  length_stops->val[i]=block.length_first(i);
            
                unsigned n_survivors=0;
                for (BlockElt z=0; z<block.size(); ++z)
                  if (block.survives(z))
                    ++n_survivors;
                own_vector survivor =
                  std::make_shared<vector_value>(int_Vector(n_survivors));
                { unsigned i=0;
                  for (BlockElt z=0; z<block.size(); ++z)
                    if (block.survives(z))
                      survivor->val[i++]=z;
                  assert(i==n_survivors);
                }
                own_matrix contributes_to = std::make_shared<matrix_value>(
                  int_Matrix(n_survivors,block.size(),0));
                for (BlockElt z=0; z<block.size(); ++z)
                { BlockEltList sb = block.survivors_below(z);
                  for (BlockEltList::const_iterator it=sb.begin(); it!=sb.end(); ++it)
                  { BlockElt x= permutations::find_index<int>(survivor->val,*it);
                      // a row index
                    if ((block.length(z)-block.length(*it))%2==0)
                      ++contributes_to->val(x,z);
                    else
                      --contributes_to->val(x,z);
                  }
                }
            
                push_value(std::move(M));
                push_value(std::move(polys));
                push_value(std::move(length_stops));
                push_value(std::move(survivor));
                push_value(std::move(contributes_to));
            
                if (l==expression_base::single_value)
                  wrap_tuple<6>();
              }
            }
            
            void split_plus_wrapper(expression_base::level l)
            { Split_integer j=get<split_int_value>()->val;
              own_split_int i=get_own<split_int_value>();
              if (l!=expression_base::no_value)
              { i->val+=j; push_value(i); }
            }
            
            void split_minus_wrapper(expression_base::level l)
            { Split_integer j=get<split_int_value>()->val;
              own_split_int i=get_own<split_int_value>();
              if (l!=expression_base::no_value)
              { i->val-=j; push_value(i); }
            }
            
            void split_unary_minus_wrapper(expression_base::level l)
            { own_split_int i=get_own<split_int_value>();
              if (l!=expression_base::no_value)
              { i->val.negate(); push_value(i); }
            }
            
            void split_times_wrapper(expression_base::level l)
            { Split_integer j=get<split_int_value>()->val;
              Split_integer i=get<split_int_value>()->val;
              if (l!=expression_base::no_value)
                push_value(std::make_shared<split_int_value>(i*j));
            }
            
            void int_to_split_coercion()
            { int a=get<int_value>()->val;
              push_value(std::make_shared<split_int_value>(Split_integer(a)));
            }
            
            void pair_to_split_coercion()
            { push_tuple_components();
              int b=get<int_value>()->val;
              int a=get<int_value>()->val;
              push_value(std::make_shared<split_int_value>(Split_integer(a,b)));
            }
            
            void from_split_wrapper(expression_base::level l)
            { Split_integer si = get<split_int_value>()->val;
              if (l!=expression_base::no_value)
              {  push_value(std::make_shared<int_value>(si.e()));
                 push_value(std::make_shared<int_value>(si.s()));
                 if (l==expression_base::single_value)
                   wrap_tuple<2>();
              }
            }
            
            void virtual_module_wrapper(expression_base::level l)
            { own_real_form rf = non_const_get<real_form_value>();
              if (l!=expression_base::no_value)
                push_value(std::make_shared<virtual_module_value> 
                  (rf,repr::SR_poly(rf->rc().repr_less())));
            }
            
            void real_form_of_virtual_module_wrapper(expression_base::level l)
            { shared_virtual_module m = get<virtual_module_value>();
              if (l!=expression_base::no_value)
                push_value(m->rf);
            }
            
            void virtual_module_size_wrapper(expression_base::level l)
            { shared_virtual_module m = get<virtual_module_value>();
              if (l!=expression_base::no_value)
                push_value(std::make_shared<int_value>(m->val.size()));
            }
            
            void param_to_poly()
            { shared_module_parameter p = get<module_parameter_value>();
            test_standard(*p);
              const own_real_form& rf=p->rf;
              push_value(std::make_shared<virtual_module_value> 
                (rf,rf->rc().expand_final(p->val)));
            }
            
            void add_module_wrapper(expression_base::level l)
            { shared_module_parameter p = get<module_parameter_value>();
              own_virtual_module accumulator = get_own<virtual_module_value>();
            test_standard(*p);
              if (accumulator->rf!=p->rf)
                throw std::runtime_error 
                  ("Real form mismatch when adding standard module to a module");
              if (l!=expression_base::no_value)
              { accumulator->val+= p->rc().expand_final(p->val);
                push_value(accumulator);
              }
            }
            
            void add_module_term_wrapper(expression_base::level l)
            { push_tuple_components(); // second argument is a pair |(coef,p)|
              own_module_parameter p = get_own<module_parameter_value>();
              Split_integer coef=get<split_int_value>()->val;
              own_virtual_module accumulator = get_own<virtual_module_value>();
            test_standard(*p);
              if (accumulator->rf!=p->rf)
                throw std::runtime_error 
                  ("Real form mismatch when adding a term to a module");
              if (l!=expression_base::no_value)
              { accumulator->val.add_multiple(p->rc().expand_final(p->val),coef);
                push_value(accumulator);
              }
            }
            
            void add_module_termlist_wrapper(expression_base::level l)
            { shared_row r = get<row_value>();
              own_virtual_module accumulator = get_own<virtual_module_value>();
              if (l!=expression_base::no_value)
              { for (auto it=r->val.cbegin(); it!=r->val.cend(); ++it)
                { const tuple_value* t = force<tuple_value>(it->get());
                  Split_integer coef=force<split_int_value>(t->val[0].get())->val;
                  const module_parameter_value* p =
                    force<module_parameter_value>(t->val[1].get());
                test_standard(*p);
                  if (accumulator->rf!=p->rf)
                    throw std::runtime_error 
                      ("Real form mismatch when adding terms to a module");
                  accumulator->val.add_multiple(p->rc().expand_final(p->val),coef);
                 }
                push_value(accumulator);
              }
            }
            
            void add_virtual_modules_wrapper(expression_base::level l)
            { own_virtual_module accumulator = get_own<virtual_module_value>();
              shared_virtual_module addend = get<virtual_module_value>();
              if (accumulator->rf!=addend->rf)
                throw std::runtime_error ("Real form mismatch when adding two modules");
              if (l!=expression_base::no_value)
              { for (repr::SR_poly::const_iterator
                     it=addend->val.begin(); it!=addend->val.end(); ++it)
                  accumulator->val.add_term(it->first,it->second);
                push_value(accumulator);
              }
            }
            
            void int_mult_virtual_module_wrapper(expression_base::level l)
            { own_virtual_module m = get_own<virtual_module_value>();
              int c = get<int_value>()->val;
              if (l!=expression_base::no_value)
              { if (c==0)
                  m->val.clear(); // avoid creating null terms into an |SR_poly|
                else
                  for (repr::SR_poly::iterator it=m->val.begin(); it!=m->val.end(); ++it)
                    it->second *= c;
                push_value(m);
              }
            }
            
            void split_mult_virtual_module_wrapper(expression_base::level l)
            { own_virtual_module m = get_own<virtual_module_value>();
              Split_integer c = get<split_int_value>()->val;
              if (l!=expression_base::no_value)
              { for (repr::SR_poly::iterator it=m->val.begin(); it!=m->val.end(); )
                  // no |++it| here!
                  if ((it->second *= c)==Split_integer(0,0))
            	m->val.erase(it++); // advance, then delete the node just abandoned
                  else ++it;
                push_value(m);
              }
            }
            
            void deform_wrapper(expression_base::level l)
            { shared_module_parameter p = get<module_parameter_value>();
              test_standard(*p);
              if (l!=expression_base::no_value)
              {
                non_integral_block block(p->rc(),p->val); // partial block construction
                repr::SR_poly terms
                   = p->rt().deformation_terms(block,block.size()-1);
            
                own_virtual_module acc = std::make_shared<virtual_module_value>
                  (p->rf, repr::SR_poly(p->rc().repr_less()));
                for (repr::SR_poly::const_iterator it=terms.begin(); it!=terms.end(); ++it)
                  acc->val.add_multiple(p->rc().expand_final(it->first),it->second);
            
                push_value(std::move(acc));
              }
            }
            
            void full_deform_wrapper(expression_base::level l)
            { shared_module_parameter p = get<module_parameter_value>();
              test_standard(*p);
              if (l!=expression_base::no_value)
              {
                repr::SR_poly result = p->rt().deformation(p->val);
                push_value(std::make_shared<virtual_module_value>(p->rf,result));
              }
            }
            
            void KL_sum_at_s_wrapper(expression_base::level l)
            { shared_module_parameter p = get<module_parameter_value>();
              test_standard(*p);
              test_nonzero_final(*p);
              if (l!=expression_base::no_value)
              {
                repr::SR_poly result = p->rt().KL_column_at_s(p->val);
                push_value(std::make_shared<virtual_module_value>(p->rf,result));
              }
            }
            
            void raw_KL_wrapper (expression_base::level l)
            { shared_Block b = get<Block_value>();
              const Block& block = b->val;
              kl::KLContext klc(block); klc.fill();
            
              if (l==expression_base::no_value)
                return;
              own_matrix M = std::make_shared<matrix_value>(int_Matrix(klc.size()));
              for (size_t y=1; y<klc.size(); ++y)
                for (size_t x=0; x<y; ++x)
                  M->val(x,y) = klc.KL_pol_index(x,y);
            
              own_row polys = std::make_shared<row_value>(0);
              polys->val.reserve(klc.polStore().size());
              for (size_t i=0; i<klc.polStore().size(); ++i)
              {
                const kl::KLPol& pol = klc.polStore()[i];
                std::vector<int> coeffs(pol.size());
                for (size_t j=pol.size(); j-->0; )
                  coeffs[j]=pol[j];
                polys->val.emplace_back(std::make_shared<vector_value>(coeffs));
              }
            
              std::vector<int> length_stops(block.length(block.size()-1)+1);
              length_stops[0]=0;
              for (size_t i=1; i<length_stops.size(); ++i)
                length_stops[i]=block.length_first(i);
            
              push_value(std::move(M));
              push_value(std::move(polys));
              push_value(std::make_shared<vector_value>(length_stops));
              if (l==expression_base::single_value)
                wrap_tuple<3>();
            }
            
            void raw_dual_KL_wrapper (expression_base::level l)
            { shared_Block b = get<Block_value>();
              const Block& block = b->val;
              Block dual_block = Block::build(b->dual_rf->val,b->rf->val);
            
              std::vector<BlockElt> dual=blocks::dual_map(block,dual_block);
              kl::KLContext klc(dual_block); klc.fill();
            
              if (l==expression_base::no_value)
                return;
              own_matrix M = std::make_shared<matrix_value>(int_Matrix(klc.size()));
              for (size_t y=1; y<klc.size(); ++y)
                for (size_t x=0; x<y; ++x)
                  M->val(x,y) = klc.KL_pol_index(dual[y],dual[x]);
            
              own_row polys = std::make_shared<row_value>(0);
              polys->val.reserve(klc.polStore().size());
              for (size_t i=0; i<klc.polStore().size(); ++i)
              {
                const kl::KLPol& pol = klc.polStore()[i];
                std::vector<int> coeffs(pol.size());
                for (size_t j=pol.size(); j-->0; )
                  coeffs[j]=pol[j];
                polys->val.emplace_back(std::make_shared<vector_value>(coeffs));
              }
            
              std::vector<int> length_stops(block.length(block.size()-1)+1);
              length_stops[0]=0;
              for (size_t i=1; i<length_stops.size(); ++i)
                length_stops[i]=block.length_first(i);
            
              push_value(std::move(M));
              push_value(std::move(polys));
              push_value(std::make_shared<vector_value>(length_stops));
              if (l==expression_base::single_value)
                wrap_tuple<3>();
            }
            
            void print_realweyl_wrapper(expression_base::level l)
            { shared_Cartan_class cc(get<Cartan_class_value>());
              own_real_form rf= non_const_get<real_form_value>();
            
              if (&rf->parent.val!=&cc->parent.val)
                throw std::runtime_error 
                ("Inner class mismatch between arguments");
            
              BitMap b(rf->parent.val.Cartan_set(rf->val.realForm()));
              if (not b.isMember(cc->number))
                throw std::runtime_error 
                ("Cartan class not defined for real form");
            
            
              realredgp_io::printRealWeyl (*output_stream,rf->val,cc->number);
            
              if (l==expression_base::single_value)
                wrap_tuple<0>();
            }
            
            
            void print_strongreal_wrapper(expression_base::level l)
            { shared_Cartan_class cc(get<Cartan_class_value>());
            
             realredgp_io::printStrongReal
                (*output_stream,cc->parent.val,cc->parent.interface,cc->number);
            
              if (l==expression_base::single_value)
                wrap_tuple<0>();
            }
            
            void print_block_wrapper(expression_base::level l)
            { shared_Block b = get<Block_value>();
              const Block& block = b->val;
            
              block.print_to(*output_stream,false);
            
              if (l==expression_base::single_value)
                wrap_tuple<0>();
            }
            
            void print_blockd_wrapper(expression_base::level l)
            { shared_Block b = get<Block_value>();
              const Block& block = b->val;
            
              block.print_to(*output_stream,true);
            
              if (l==expression_base::single_value)
                wrap_tuple<0>();
            }
            
            
            void print_blocku_wrapper(expression_base::level l)
            { shared_Block b = get<Block_value>();
              const Block& block = b->val;
            
              block_io::printBlockU(*output_stream,block);
            
              if (l==expression_base::single_value)
                wrap_tuple<0>();
            }
            
            void print_blockstabilizer_wrapper(expression_base::level l)
            { shared_Cartan_class cc(get<Cartan_class_value>());
              shared_Block b = get<Block_value>();
            
              realredgp_io::printBlockStabilizer
               (*output_stream,b->rf->val,cc->number,b->dual_rf->val.realForm());
            
              if (l==expression_base::single_value)
                wrap_tuple<0>();
            }
            
            void print_KGB_wrapper(expression_base::level l)
            { own_real_form rf= non_const_get<real_form_value>();
            
              *output_stream
                << "kgbsize: " << rf->val.KGB_size() << std::endl;
              const KGB& kgb=rf->kgb();
              kgb_io::var_print_KGB(*output_stream,rf->val.complexGroup(),kgb);
            
              if (l==expression_base::single_value)
                wrap_tuple<0>();
            }
            
            void print_X_wrapper(expression_base::level l)
            { shared_inner_class ic = get<inner_class_value>();
            
              ComplexReductiveGroup& G=ic->val;
              kgb::global_KGB kgb(G); // build global Tits group, "all" square classes
              kgb_io::print_X(*output_stream,kgb);
            
              if (l==expression_base::single_value)
                wrap_tuple<0>();
            }
            
            void print_KL_basis_wrapper(expression_base::level l)
            { own_Block b = non_const_get<Block_value>();
              Block& block = b->val;
              kl::KLContext klc(block); klc.fill(false);
            
              *output_stream
                << "Full list of non-zero Kazhdan-Lusztig-Vogan polynomials:\n\n";
              kl_io::printAllKL(*output_stream,klc,block);
            
              if (l==expression_base::single_value)
                wrap_tuple<0>();
            }
            
            void print_prim_KL_wrapper(expression_base::level l)
            { own_Block b = non_const_get<Block_value>();
              Block &block = b->val; // this one must be non-|const|
              kl::KLContext klc(block); klc.fill(false);
            
              *output_stream
                << "Non-zero Kazhdan-Lusztig-Vogan polynomials for primitive pairs:\n\n";
              kl_io::printPrimitiveKL(*output_stream,klc,block);
            
              if (l==expression_base::single_value)
                wrap_tuple<0>();
            }
            
            void print_KL_list_wrapper(expression_base::level l)
            { shared_Block b = get<Block_value>();
              const Block &block = b->val;
              kl::KLContext klc(block); klc.fill(false);
            
              kl_io::printKLList(*output_stream,klc);
            
              if (l==expression_base::single_value)
                wrap_tuple<0>();
            }
            
            void print_W_cells_wrapper(expression_base::level l)
            { shared_Block b = get<Block_value>();
              const Block &block = b->val;
              kl::KLContext klc(block); klc.fill(false);
            
            
              wgraph::WGraph wg(klc.rank()); kl::wGraph(wg,klc);
              wgraph::DecomposedWGraph dg(wg);
            
              wgraph_io::printWDecomposition(*output_stream,dg);
            
              if (l==expression_base::single_value)
                wrap_tuple<0>();
            }
            
            void print_W_graph_wrapper(expression_base::level l)
            { shared_Block b = get<Block_value>();
              const Block &block = b->val;
              kl::KLContext klc(block); klc.fill(false);
            
            
              wgraph::WGraph wg(klc.rank()); kl::wGraph(wg,klc);
            
              wgraph_io::printWGraph(*output_stream,wg);
            
              if (l==expression_base::single_value)
                wrap_tuple<0>();
            }
 }

void initialise_builtin_types()
{ 
  {
    coercion(str_type,Lie_type_type,"LT",Lie_type_coercion);
    coercion(ic_type,rd_type,"RdIc",inner_class_to_root_datum_coercion);
    coercion(rf_type,ic_type,"IcRf",real_form_to_inner_class_coercion);
    coercion(rf_type,rd_type,"RdRf",real_form_to_root_datum_coercion);
    coercion(int_type,split_type,"SpI",int_to_split_coercion);
    coercion(int_int_type,split_type,"Sp(I,I)",pair_to_split_coercion);
    coercion(param_type,param_pol_type,"PolP",param_to_poly);
  }
  
  install_function(Lie_type_wrapper,"Lie_type","(string->LieType)");
  install_function(Cartan_matrix_wrapper,"Cartan_matrix","(LieType->mat)");
  install_function(type_of_Cartan_matrix_wrapper
  		,"Cartan_matrix_type","(mat->LieType,vec)");
  install_function(Lie_rank_wrapper,"rank","(LieType->int)");
  install_function(semisimple_rank_wrapper,"semisimple_rank","(LieType->int)");
  install_function(Lie_type_string_wrapper,"str","(LieType->string)");
  install_function(nr_factors_wrapper,"#","(LieType->int)");
  install_function(Lie_factors_wrapper,"%","(LieType->[LieType])");
  
  install_function(Smith_Cartan_wrapper,"Smith_Cartan","(LieType->mat,vec)");
  install_function(filter_units_wrapper,"filter_units","(mat,vec->mat,vec)");
  install_function(ann_mod_wrapper,"ann_mod","(mat,int->mat)");
  install_function(replace_gen_wrapper,"replace_gen",
  		"((mat,vec),mat->mat)");
  install_function(quotient_basis_wrapper
  		,"quotient_basis","(LieType,[ratvec]->mat)");
  install_function(basic_involution_wrapper,"involution",
  		"(LieType,string->mat)");
  install_function(based_involution_wrapper,"involution",
  		"(LieType,mat,string->mat)");
  
  install_function(type_of_root_datum_wrapper,"Lie_type"
                  ,"(RootDatum->LieType)");
  install_function(raw_root_datum_wrapper,"root_datum"
                  ,"(mat,mat->RootDatum)");
  install_function(root_datum_wrapper,"root_datum","(LieType,mat->RootDatum)");
  install_function(sublattice_root_datum_wrapper,"root_datum"
                  ,"(RootDatum,mat->RootDatum)");
  install_function(quotient_datum_wrapper
  		,"root_datum","(LieType,[ratvec]->RootDatum)");
  install_function(simply_connected_datum_wrapper
  		,"simply_connected","(LieType->RootDatum)");
  install_function(adjoint_datum_wrapper, "adjoint","(LieType->RootDatum)");
  install_function(simple_roots_wrapper,"simple_roots","(RootDatum->mat)");
  install_function(simple_coroots_wrapper,"simple_coroots","(RootDatum->mat)");
  install_function(positive_roots_wrapper, "posroots","(RootDatum->mat)");
  install_function(positive_coroots_wrapper, "poscoroots","(RootDatum->mat)");
  install_function(datum_Cartan_wrapper,"Cartan_matrix","(RootDatum->mat)");
  install_function(roots_wrapper,"roots","(RootDatum->mat)");
  install_function(coroots_wrapper,"coroots","(RootDatum->mat)");
  install_function(root_coradical_wrapper,"root_coradical","(RootDatum->mat)");
  install_function(coroot_radical_wrapper,"coroot_radical","(RootDatum->mat)");
  install_function(fundamental_weight_wrapper,
  		 "fundamental_weight","(RootDatum,int->ratvec)");
  install_function(fundamental_coweight_wrapper,
  		 "fundamental_coweight","(RootDatum,int->ratvec)");
  install_function(dual_datum_wrapper,"dual","(RootDatum->RootDatum)");
  install_function(derived_info_wrapper,
  		 "derived_info","(RootDatum->RootDatum,mat)");
  install_function(mod_central_torus_info_wrapper,
  		 "mod_central_torus_info","(RootDatum->RootDatum,mat)");
  install_function(rd_rank_wrapper,"rank","(RootDatum->int)");
  install_function(rd_semisimple_rank_wrapper
  		,"semisimple_rank","(RootDatum->int)");
  install_function(rd_nposroots_wrapper
  		,"nr_of_posroots","(RootDatum->int)");
  install_function(integrality_datum_wrapper
                  ,"integrality_datum","(RootDatum,ratvec->RootDatum)");
  install_function(integrality_points_wrapper
                  ,"integrality_points","(RootDatum,ratvec->[rat])");
  
  install_function(classify_wrapper,"classify_involution"
                  ,"(mat->int,int,int)");
  install_function(fix_involution_wrapper,"inner_class"
                  ,"(RootDatum,mat->InnerClass)");
  install_function(twisted_involution_wrapper,"twisted_involution"
                  ,"(RootDatum,mat->InnerClass,vec)");
  install_function(set_type_wrapper,"inner_class"
                  ,"(LieType,[ratvec],string->InnerClass)");
  install_function(set_inner_class_wrapper,"inner_class"
                  ,"(RootDatum,string->InnerClass)");
  install_function(inner_class_equality_wrapper,"="
                  ,"(InnerClass,InnerClass->bool)");
  install_function(distinguished_involution_wrapper,"distinguished_involution"
                  ,"(InnerClass->mat)");
  install_function(root_datum_of_inner_class_wrapper,"root_datum"
                  ,"(InnerClass->RootDatum)");
  install_function(dual_inner_class_wrapper,"dual"
                  ,"(InnerClass->InnerClass)");
  install_function(form_names_wrapper,"form_names"
                  ,"(InnerClass->[string])");
  install_function(dual_form_names_wrapper,"dual_form_names"
                  ,"(InnerClass->[string])");
  install_function(n_real_forms_wrapper,"nr_of_real_forms"
                  ,"(InnerClass->int)");
  install_function(n_dual_real_forms_wrapper,"nr_of_dual_real_forms"
                  ,"(InnerClass->int)");
  install_function(n_Cartan_classes_wrapper,"nr_of_Cartan_classes"
                  ,"(InnerClass->int)");
  install_function(block_sizes_wrapper,"block_sizes"
                  ,"(InnerClass->mat)");
  install_function(occurrence_matrix_wrapper,"occurrence_matrix"
                  ,"(InnerClass->mat)");
  install_function(dual_occurrence_matrix_wrapper,"dual_occurrence_matrix"
                  ,"(InnerClass->mat)");
  
  install_function(real_form_wrapper,"real_form","(InnerClass,int->RealForm)");
  install_function(form_number_wrapper,"form_number","(RealForm->int)");
  install_function(quasisplit_form_wrapper,"quasisplit_form"
  		,"(InnerClass->RealForm)");
  install_function(inner_class_of_real_form_wrapper
                  ,"inner_class","(RealForm->InnerClass)");
  install_function(components_rank_wrapper,"components_rank","(RealForm->int)");
  install_function(count_Cartans_wrapper,"count_Cartans","(RealForm->int)");
  install_function(KGB_size_wrapper,"KGB_size","(RealForm->int)");
  install_function(base_grading_vector_wrapper
                  ,"base_grading_vector","(RealForm->ratvec)");
  install_function(Cartan_order_matrix_wrapper,"Cartan_order","(RealForm->mat)");
  install_function(real_form_equals_wrapper,"=","(RealForm,RealForm->bool)");
  install_function(dual_real_form_wrapper,"dual_real_form"
  				       ,"(InnerClass,int->RealForm)");
  install_function(dual_quasisplit_form_wrapper,"dual_quasisplit_form"
  		,"(InnerClass->RealForm)");
  
  install_function(ic_Cartan_class_wrapper,"Cartan_class"
  		,"(InnerClass,int->CartanClass)");
  install_function(rf_Cartan_class_wrapper,"Cartan_class"
  		,"(RealForm,int->CartanClass)");
  install_function(most_split_Cartan_wrapper,"most_split_Cartan"
  		,"(RealForm->CartanClass)");
  install_function(Cartan_class_real_form_wrapper,"Cartan_class_real_form"
  		,"(InnerClass,mat,ratvec->CartanClass,RealForm)");
  install_function(Cartan_involution_wrapper,"involution","(CartanClass->mat)");
  install_function(Cartan_info_wrapper,"Cartan_info"
  		,"(CartanClass->(int,int,int),"
                   "vec,(int,int),(LieType,LieType,LieType))");
  install_function(real_forms_of_Cartan_wrapper,"real_forms"
  		,"(CartanClass->[RealForm])");
  install_function(dual_real_forms_of_Cartan_wrapper,"dual_real_forms"
  		,"(CartanClass->[RealForm])");
  install_function(fiber_partition_wrapper,"fiber_partition"
  		,"(CartanClass,RealForm->[int])");
  install_function(square_classes_wrapper,"square_classes"
                  ,"(CartanClass->[[int]])");
  
  install_function(KGB_elt_wrapper,"KGB","(RealForm,int->KGBElt)");
  install_function(decompose_KGB_wrapper,"%","(KGBElt->RealForm,int)");
  install_function(KGB_cross_wrapper,"cross","(int,KGBElt->KGBElt)");
  install_function(KGB_Cayley_wrapper,"Cayley","(int,KGBElt->KGBElt)");
  install_function(KGB_inv_Cayley_wrapper,"inv_Cayley","(int,KGBElt->KGBElt)");
  install_function(KGB_status_wrapper,"status","(int,KGBElt->int)");
  install_function(root_status_wrapper,"status","(vec,KGBElt->int)");
  install_function(build_KGB_element_wrapper,"KGB_elt"
  		,"(RealForm,mat,ratvec->KGBElt)");
  install_function(KGB_twist_wrapper,"twist","(KGBElt->KGBElt)");
  install_function(KGB_Cartan_wrapper,"Cartan_class","(KGBElt->CartanClass)");
  install_function(KGB_involution_wrapper,"involution","(KGBElt->mat)");
  install_function(KGB_length_wrapper,"length","(KGBElt->int)");
  install_function(torus_bits_wrapper,"torus_bits","(KGBElt->vec)");
  install_function(torus_factor_wrapper,"torus_factor","(KGBElt->ratvec)");
  install_function(KGB_equals_wrapper,"=","(KGBElt,KGBElt->bool)");
  
  install_function(Fokko_block_wrapper,"block","(RealForm,RealForm->Block)");
  install_function(decompose_block_wrapper,"%","(Block->RealForm,RealForm)");
  install_function(block_size_wrapper,"#","(Block->int)");
  install_function(block_element_wrapper,"element","(Block,int->KGBElt,KGBElt)");
  install_function(block_index_wrapper,"index","(Block,KGBElt,KGBElt->int)");
  install_function(dual_block_wrapper,"dual","(Block->Block)");
  install_function(block_status_wrapper,"status","(int,Block,int->int)");
  install_function(block_cross_wrapper,"cross","(int,Block,int->int)");
  install_function(block_Cayley_wrapper,"Cayley","(int,Block,int->int)");
  install_function(block_inverse_Cayley_wrapper,"inverse_Cayley"
                  ,"(int,Block,int->int)");
  
  install_function(module_parameter_wrapper,"param"
                  ,"(KGBElt,vec,ratvec->Param)");
  install_function(unwrap_parameter_wrapper,"%"
                  ,"(Param->KGBElt,vec,ratvec)");
  install_function(infinitesimal_character_wrapper,"infinitesimal_character"
                  ,"(Param->ratvec)");
  install_function(real_form_of_parameter_wrapper,"real_form"
  		,"(Param->RealForm)");
  install_function(is_standard_wrapper,"is_standard" ,"(Param->bool)");
  install_function(is_zero_wrapper,"is_zero" ,"(Param->bool)");
  install_function(is_final_wrapper,"is_final" ,"(Param->bool)");
  install_function(parameter_dominant_wrapper,"dominant" ,"(Param->Param)");
  install_function(parameter_equivalent_wrapper,"=" ,"(Param,Param->bool)");
  install_function(parameter_cross_wrapper,"cross" ,"(int,Param->Param)");
  install_function(parameter_Cayley_wrapper,"Cayley" ,"(int,Param->Param)");
  install_function(parameter_inv_Cayley_wrapper,"inv_Cayley"
                  ,"(int,Param->Param)");
  install_function(parameter_twist_wrapper,"twist" ,"(Param->Param)");
  install_function(orientation_number_wrapper,"orientation_nr" ,"(Param->int)");
  install_function(reducibility_points_wrapper,
  		"reducibility_points" ,"(Param->[rat])");
  install_function(print_n_block_wrapper,"print_block","(Param->)");
  install_function(block_wrapper,"block" ,"(Param->[Param],int)");
  install_function(partial_block_wrapper,"partial_block","(Param->[Param])");
  install_function(param_length_wrapper,"length","(Param->int)");
  install_function(KL_block_wrapper,"KL_block"
                  ,"(Param->[Param],int,mat,[vec],vec,vec,mat)");
  install_function(partial_KL_block_wrapper,"partial_KL_block"
                  ,"(Param->[Param],mat,[vec],vec,vec,mat)");
  
  install_function(split_plus_wrapper,"+","(Split,Split->Split)");
  install_function(split_minus_wrapper,"-","(Split,Split->Split)");
  install_function(split_unary_minus_wrapper,"-","(Split->Split)");
  install_function(split_times_wrapper,"*","(Split,Split->Split)");
  install_function(from_split_wrapper,"%","(Split->int,int)");
  install_function(virtual_module_wrapper,"null_module","(RealForm->ParamPol)");
  install_function(real_form_of_virtual_module_wrapper,"real_form"
  		,"(ParamPol->RealForm)");
  install_function(virtual_module_size_wrapper,"#","(ParamPol->int)");
  install_function(add_module_wrapper,"+","(ParamPol,Param->ParamPol)");
  install_function(add_module_term_wrapper,"+"
  		,"(ParamPol,(Split,Param)->ParamPol)");
  install_function(add_module_termlist_wrapper,"+"
  		,"(ParamPol,[(Split,Param)]->ParamPol)");
  install_function(add_virtual_modules_wrapper,"+"
  		,"(ParamPol,ParamPol->ParamPol)");
  install_function(int_mult_virtual_module_wrapper,"*"
  		,"(int,ParamPol->ParamPol)");
  install_function(split_mult_virtual_module_wrapper,"*"
  		,"(Split,ParamPol->ParamPol)");
  install_function(deform_wrapper,"deform" ,"(Param->ParamPol)");
  install_function(full_deform_wrapper,"full_deform","(Param->ParamPol)");
  install_function(KL_sum_at_s_wrapper,"KL_sum_at_s","(Param->ParamPol)");
  
  install_function(raw_KL_wrapper,"raw_KL","(Block->mat,[vec],vec)");
  install_function(raw_dual_KL_wrapper,"dual_KL","(Block->mat,[vec],vec)");
  install_function(print_gradings_wrapper,"print_gradings"
  		,"(CartanClass,RealForm->)");
  install_function(print_realweyl_wrapper,"print_real_Weyl"
  		,"(RealForm,CartanClass->)");
  install_function(print_strongreal_wrapper,"print_strong_real"
  		,"(CartanClass->)");
  install_function(print_block_wrapper,"print_block","(Block->)");
  install_function(print_blocku_wrapper,"print_blocku","(Block->)");
  install_function(print_blockd_wrapper,"print_blockd","(Block->)");
  install_function(print_blockstabilizer_wrapper,"print_blockstabilizer"
  		,"(Block,CartanClass->)");
  install_function(print_KGB_wrapper,"print_KGB","(RealForm->)");
  install_function(print_X_wrapper,"print_X","(InnerClass->)");
  install_function(print_KL_basis_wrapper,"print_KL_basis","(Block->)");
  install_function(print_prim_KL_wrapper,"print_prim_KL","(Block->)");
  install_function(print_KL_list_wrapper,"print_KL_list","(Block->)");
  install_function(print_W_cells_wrapper,"print_W_cells","(Block->)");
  install_function(print_W_graph_wrapper,"print_W_graph","(Block->)");
}

void Lie_type_value::add_simple_factor (char c,size_t rank)
   throw(std::bad_alloc,std::runtime_error)
{ static const std::string types=lietype::typeLetters; // |"ABCDEFGT"|
  size_t t=types.find(c);
  if (t==std::string::npos)
    throw std::runtime_error(std::string("Invalid type letter '")+c+'\'');

  const size_t r=constants::RANK_MAX;
static const size_t lwb[]={1,2,2,4,6,4,2,0};
  static const size_t upb[]={r,r,r,r,8,4,2,r};
  if (rank<lwb[t])
    throw std::runtime_error("Too small rank "+str(rank)+" for Lie type "+c);

  if (rank>upb[t])
    if (upb[t]!=r)
      throw std::runtime_error("Too large rank "+str(rank)+" for Lie type "+c);

    else
      throw std::runtime_error 
      ("Rank "+str(rank)+" exceeds implementation limit "+str(r));

  if (c=='T')
    while (rank-->0) val.push_back(SimpleLieType('T',1));
  else
    val.push_back(SimpleLieType(c,rank));
}

void Lie_type_value::print(std::ostream& out) const
{ if (val.empty()) out << "empty Lie type";
  else
    out << "Lie type '" << val << '\'';
}

void root_datum_value::print(std::ostream& out) const
{ Lie_type_value type(type_of_datum(val));
  out << (val.isSimplyConnected() ? "simply connected " : "")
    << (val.isAdjoint() ? "adjoint " : "")
      << "root datum of " << type;
}

inner_class_value::inner_class_value(const inner_class_value& v)
: val(v.val), dual(v.dual), ref_count(v.ref_count)
, rd_type(v.rd_type), ic_type(v.ic_type)
, interface(v.interface), dual_interface(v.dual_interface)
{ ++ref_count; }

inner_class_value::~inner_class_value()
{ if (--ref_count==0) { delete &val; delete &dual; delete &ref_count;} }

inner_class_value::inner_class_value
  (std::unique_ptr<ComplexReductiveGroup> g,
   const lietype::Layout& lo)
: val(*g)
, dual(*new ComplexReductiveGroup(*g,tags::DualTag()))
, ref_count(*new size_t(1))
, rd_type(lo.d_type), ic_type(lo.d_inner)
, interface(*g,lo), dual_interface(*g,lo,tags::DualTag())
 { g.release(); } 

inner_class_value::inner_class_value(const inner_class_value& v,tags::DualTag)
: val(v.dual), dual(v.val)
, ref_count(v.ref_count)
, rd_type(lietype::dual_type(v.rd_type))
, ic_type(lietype::dual_type(v.ic_type,rd_type))
, interface(v.dual_interface), dual_interface(v.interface)
{ ++ref_count; }

void inner_class_value::print(std::ostream& out) const
{ out << "Complex reductive group of type " << rd_type 
      << ", with involution defining\n"
         "inner class of type '" << ic_type 
      << "', with " << val.numRealForms()  << " real "
      << (val.numRealForms()==1 ? "form" : "forms")  << " and "
      << val.numDualRealForms()  << " dual real "
      << (val.numDualRealForms()==1 ? "form" : "forms");
}

  const Rep_context& real_form_value::rc()
    { return *(rt_p==NULL ? rt_p=new Rep_table(val) : rt_p); }
  Rep_table& real_form_value::rt()
    { return *(rt_p==NULL ? rt_p=new Rep_table(val) : rt_p); }

void real_form_value::print(std::ostream& out) const
{ if (val.isCompact()) out << "compact ";
  out << (val.isConnected() ? "connected " : "disconnected " );
  if (val.isQuasisplit())
    out << (val.isSplit() ? "" : "quasi") << "split ";
  out << "real group with Lie algebra '" 
      << parent.interface.typeName(parent.interface.out(val.realForm())) 
      << '\'' ;
}

Cartan_class_value::Cartan_class_value(const inner_class_value& p,size_t cn)
: parent(p),number(cn),val(p.val.cartan(cn))
{}

void Cartan_class_value::print(std::ostream& out) const
{ out << "Cartan class #" << number << ", occurring for " 
  << val.numRealForms() << " real " 
  << (val.numRealForms()==1 ? "form" : "forms") << " and for "
  << val.numDualRealForms() << " dual real "  
  << (val.numDualRealForms()==1 ? "form" : "forms");
}

void KGB_elt_value::print(std::ostream& out) const
{ out << "KGB element #" << val;
}

  Block_value::Block_value(const own_real_form& form,
                          const own_real_form& dual_form)
  : rf(form), dual_rf(dual_form)
  ,
  val(Block::build(rf->val,dual_rf->val)) {}

void Block_value::print(std::ostream& out) const
{ out << "Block of " << val.size() << " elements";
}

std::ostream& print
  (std::ostream& out,const StandardRepr& val, const Rep_context& rc)
{ RootNbr witness; // dummy needed in call
  return
  out << 
       ( rc.is_standard(val,witness) ?
           rc.is_zero(val,witness) ? "zero" :
             rc.is_final(val,witness) ? "final" : "non-final"
         : "non-standard" )

    << " parameter (x="
      << val.x() << ",lambda="
      << rc.lambda(val) << ",nu="
      << rc.nu(val) << ')';
}

void module_parameter_value::print(std::ostream& out) const
{ interpreter::print(out,val,rc()); }

std::ostream& print (std::ostream& out, const Split_integer& val)
{
  return out << '(' << val.e() << (val.s()<0?'-':'+') << abs(val.s()) << "s)";
}

void split_int_value::print(std::ostream& out) const 
{ interpreter::print(out,val); }

void virtual_module_value::print(std::ostream& out) const
{ if (val.empty())
    { out << "Empty sum of standard modules"; return; }
  bool has_one=false, has_s=false;
  for (repr::SR_poly::const_iterator it=val.begin(); it!=val.end(); ++it)
  { if (it->second.e()!=0)
      has_one=true;
    if (it->second.s()!=0)
      has_s=true;
    if (has_one and has_s)
      break;
  }
  assert (has_one or has_s); // otherwise the module would have been empty
  for (repr::SR_poly::const_iterator it=val.begin(); it!=val.end(); ++it)
  { out << '\n';
    if (has_one and has_s)
      interpreter::print(out,it->second); // print coefficient
    else if (has_one)
      out << it->second.e();
    else
      out << it->second.s() << 's';
    interpreter::print(out << '*',it->first,rc()); // print parameter
  }
}

void module_coefficient::evaluate(level l) const
{ shared_virtual_module m = (array->eval(),get<virtual_module_value>());
  shared_module_parameter p = (index->eval(),get<module_parameter_value>());
  if (l!=expression_base::no_value)
    push_value(std::make_shared<split_int_value>(m->val[p->val]));
}

} }

