
#ifndef LEXER_H
#define LEXER_H

#include  <cstdlib>
#include  "buffer.h"
#include  "parsetree.h"
#include  "parser.tab.h"
#include <cctype>
#include <iostream>
#include  <string>

namespace atlas { namespace interpreter {


class Lexical_analyser
{ enum states  { initial, normal, ended };
BufferedInput& input;
  Hash_table& id_table;
  id_type keyword_limit; // first non-keyword identifier
  id_type type_limit; // first non-type identifier
  int nesting; // number of pending opening symbols
  char prevent_termination; // either |'\0'| or character requiring more input
  int comment_start, comment_end; // characters that start/end a comment
  states state; // to trigger special behaviour
  std::string file_name; // stores a file name for I/O redirection
public:
  Lexical_analyser
    (BufferedInput&, Hash_table&, const char**, const char** type_names);
  int get_token(YYSTYPE *valp, YYLTYPE* locp);
  bool reset(); // get clean slate, return |false| if |getline()| fails
  void set_comment_delims (char c, char d) 
          { comment_start=c; comment_end=d; }
  const char* scanned_file_name() const { return file_name.c_str(); }
private:
  void skip_space();
  std::string scan_quoted_string();
};


extern "C" char* id_completion_func(const char* text, int state);


extern Hash_table* main_hash_table;

extern char lexical_break_chars[];

extern Lexical_analyser* lex;


} }
#endif

